# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import dy, pl, plexc


class MyColSpec(cs.ColSpec):
    a = cs.Float64()
    b = cs.String()


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
@pytest.mark.parametrize(
    "data",
    [
        {"a": [3], "b": [1]},
        {"a": [1], "b": [2], "c": [3]},
    ],
)
def test_cast_valid(
    df_type: type[pl.DataFrame] | type[pl.LazyFrame], data: dict[str, Any]
):
    df = df_type(data)
    out = MyColSpec.cast_polars(df)
    assert isinstance(out, df_type)
    assert out.lazy().collect_schema() == MyColSpec.polars_schema()


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_cast_invalid_schema_eager():
    df = pl.DataFrame({"a": [1]})
    with pytest.raises(plexc.ColumnNotFoundError):
        MyColSpec.cast_polars(df)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_cast_invalid_schema_lazy():
    lf = pl.LazyFrame({"a": [1]})
    lf = MyColSpec.cast_polars(lf)
    with pytest.raises(plexc.ColumnNotFoundError):
        lf.collect()



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.columns.integer import _BaseInteger
from pydiverse.colspec.optional_dependency import (
    FLOAT_DTYPES,
    INTEGER_DTYPES,
    C,
    DataTypeClass,
    dy,
    pdt,
    pl,
)
from pydiverse.colspec.testing import INTEGER_COLUMN_TYPES
from pydiverse.colspec.testing.rules import evaluate_rules


class IntegerColSpec(cs.ColSpec):
    a = cs.Integer()


@pytest.mark.parametrize("column_type", INTEGER_COLUMN_TYPES)
@pytest.mark.parametrize(
    "kwargs",
    [
        {"min": 2, "max": 1},
        {"min_exclusive": 2, "max": 2},
        {"min": 2, "max_exclusive": 2},
        {"min_exclusive": 2, "max_exclusive": 2},
        {"min": 2, "min_exclusive": 2},
        {"max": 2, "max_exclusive": 2},
    ],
)
def test_args_consistency_min_max(
    column_type: type[_BaseInteger], kwargs: dict[str, Any]
):
    with pytest.raises(ValueError):
        column_type(**kwargs)


@pytest.mark.parametrize("column_type", INTEGER_COLUMN_TYPES)
def test_invalid_args_min_max(column_type: type[_BaseInteger]):
    with pytest.raises(ValueError):
        column_type(min=column_type.min_value - 1)
    with pytest.raises(ValueError):
        column_type(max=column_type.max_value + 1)


@pytest.mark.parametrize("column_type", INTEGER_COLUMN_TYPES)
@pytest.mark.parametrize(
    "kwargs",
    [
        {"min": 1, "is_in": [2, 3, 4]},
        {"max": 2, "is_in": [4, 5, 6]},
        {"min": 1, "max": 5, "is_in": [2, 3, 4]},
    ],
)
def test_invalid_args_is_in(column_type: type[_BaseInteger], kwargs: dict[str, Any]):
    with pytest.raises(ValueError):
        column_type(**kwargs)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("dtype", INTEGER_DTYPES)
def test_any_integer_dtype_passes_polars(dtype: DataTypeClass):
    df = pl.DataFrame(schema={"a": dtype})
    assert IntegerColSpec.is_valid_polars(df)


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("dtype", INTEGER_DTYPES)
def test_any_integer_dtype_passes(dtype: DataTypeClass):
    if dtype == pl.Int128:
        # this type is not supported by pydiverse libraries, yet
        return
    df = pl.DataFrame(schema={"a": dtype})
    tbl = pdt.Table(df)
    assert IntegerColSpec.is_valid(tbl)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("dtype", [pl.Boolean, pl.String] + list(FLOAT_DTYPES))
def test_non_integer_dtype_fails_polars(dtype: DataTypeClass):
    df = pl.DataFrame(schema={"a": dtype})
    assert not IntegerColSpec.is_valid_polars(df)


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("dtype", [pl.Boolean, pl.String] + list(FLOAT_DTYPES))
def test_non_integer_dtype_fails(dtype: DataTypeClass):
    df = pl.DataFrame(schema={"a": dtype})
    tbl = pdt.Table(df)
    assert not IntegerColSpec.is_valid(tbl)


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("column_type", INTEGER_COLUMN_TYPES)
@pytest.mark.parametrize("inclusive", [True, False])
def test_validate_min(column_type: type[_BaseInteger], inclusive: bool):
    kwargs = {("min" if inclusive else "min_exclusive"): 3}
    column = column_type(**kwargs)  # type: ignore
    tbl = pdt.Table({"a": [1, 2, 3, 4, 5]})
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    key = "min" if inclusive else "min_exclusive"
    expected = {key: [False, False, inclusive, True, True]}
    assert actual == expected


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("column_type", INTEGER_COLUMN_TYPES)
@pytest.mark.parametrize("inclusive", [True, False])
def test_validate_max(column_type: type[_BaseInteger], inclusive: bool):
    kwargs = {("max" if inclusive else "max_exclusive"): 3}
    column = column_type(**kwargs)  # type: ignore
    tbl = pdt.Table({"a": [1, 2, 3, 4, 5]})
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    key = "max" if inclusive else "max_exclusive"
    expected = {key: [True, True, inclusive, False, False]}
    assert actual == expected


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("column_type", INTEGER_COLUMN_TYPES)
@pytest.mark.parametrize("min_inclusive", [True, False])
@pytest.mark.parametrize("max_inclusive", [True, False])
def test_validate_range(
    column_type: type[_BaseInteger], min_inclusive: bool, max_inclusive: bool
):
    kwargs = {
        ("min" if min_inclusive else "min_exclusive"): 2,
        ("max" if max_inclusive else "max_exclusive"): 4,
    }
    column = column_type(**kwargs)  # type: ignore
    tbl = pdt.Table({"a": [1, 2, 3, 4, 5]})
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    key_min = "min" if min_inclusive else "min_exclusive"
    key_max = "max" if max_inclusive else "max_exclusive"
    expected = {
        key_min: [False, min_inclusive, True, True, True],
        key_max: [True, True, True, max_inclusive, False],
    }
    assert actual == expected


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("column_type", INTEGER_COLUMN_TYPES)
def test_validate_is_in(column_type: type[_BaseInteger]):
    column = column_type(is_in=[3, 5])
    tbl = pdt.Table({"a": [1, 2, 3, 4, 5]})
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    expected = {"is_in": [False, False, True, False, True]}
    assert actual == expected


@pytest.mark.parametrize(
    ("column_type", "num_bytes"),
    [
        (cs.Integer, 8),
        (cs.Int8, 1),
        (cs.Int16, 2),
        (cs.Int32, 4),
        (cs.Int64, 8),
        (cs.UInt8, 1),
        (cs.UInt16, 2),
        (cs.UInt32, 4),
        (cs.UInt64, 8),
    ],
)
def test_num_bytes(column_type: type[_BaseInteger], num_bytes: int):
    assert column_type.num_bytes == num_bytes


@pytest.mark.parametrize(
    ("column_type", "is_unsigned"),
    [
        (cs.Integer, False),
        (cs.Int8, False),
        (cs.Int16, False),
        (cs.Int32, False),
        (cs.Int64, False),
        (cs.UInt8, True),
        (cs.UInt16, True),
        (cs.UInt32, True),
        (cs.UInt64, True),
    ],
)
def test_is_unsigned(column_type: type[_BaseInteger], is_unsigned: bool):
    assert column_type.is_unsigned == is_unsigned


@pytest.mark.parametrize(
    ("column_type", "min_value", "max_value"),
    [
        (cs.Integer, -9223372036854775808, 9223372036854775807),
        (cs.Int8, -128, 127),
        (cs.Int16, -32768, 32767),
        (cs.Int32, -2147483648, 2147483647),
        (cs.Int64, -9223372036854775808, 9223372036854775807),
        (cs.UInt8, 0, 255),
        (cs.UInt16, 0, 65535),
        (cs.UInt32, 0, 4294967295),
        (cs.UInt64, 0, 18446744073709551615),
    ],
)
def test_type_min_max_values(
    column_type: type[_BaseInteger], min_value: int, max_value: int
):
    assert column_type.min_value == min_value
    assert column_type.max_value == max_value



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec import Collection, ColSpec
from pydiverse.colspec._filter import Filter
from pydiverse.colspec.exc import AnnotationImplementationError, ImplementationError
from pydiverse.colspec.optional_dependency import pl
from pydiverse.colspec.testing.factory import (
    create_collection,
    create_collection_raw,
    create_colspec,
)


def test_annotation_union_success():
    """When we use a union annotation, it must contain one typed LazyFrame and None."""
    res = create_collection_raw(
        "test",
        {
            "first": (
                create_colspec("first", {"a": cs.Integer(primary_key=True)}) | None
            ),
        },
    ).members()
    assert len(res) == 1
    assert res["first"].is_optional
    assert res["first"].col_spec.primary_keys() == ["a"]
    assert issubclass(res["first"].col_spec, ColSpec)


def test_annotation_union_too_many_arg_failure():
    """Unions should have a maximum of two types in them."""

    with pytest.raises(AnnotationImplementationError):
        create_collection_raw(
            "test",
            {
                "first": create_colspec("first", {"a": cs.Integer(primary_key=True)})
                | create_colspec("second", {"a": cs.Integer(primary_key=True)})
                | None,
            },
        ).members()


def test_annotation_union_conflicting_types_failure():
    """Unions should contain a maximum of one non-None type."""

    with pytest.raises(AnnotationImplementationError):
        create_collection_raw(
            "test",
            {
                "first": create_colspec("first", {"a": cs.Integer(primary_key=True)})
                | create_colspec("second", {"a": cs.Integer(primary_key=True)}),
            },
        ).members()


def test_annotation_only_none_failure():
    """Annotations must not just be None."""
    res = create_collection_raw(
        "test",
        {"first": None, "second": int, "third": Collection},
    ).members()
    assert len(res) == 0


def test_annotation_invalid_type_failure():
    """Annotations must not just be None."""
    res = create_collection_raw(
        "test",
        {
            "first": int | None,
        },
    ).members()
    assert len(res) == 0


def test_name_overlap():
    with pytest.raises(
        ImplementationError,
        match=r"Collection cannot have a filter named '_primary_key_'",
    ):
        coll = create_collection(
            "test",
            {
                "first": create_colspec("first", {"a": cs.Integer(primary_key=True)}),
                "second": create_colspec("second", {"a": cs.Integer(primary_key=True)}),
            },
            filters={"_primary_key_": Filter(lambda c: c.first)},
        )()
        coll.filter_rules()


def test_collection_no_primary_key_success():
    """It's ok not to have primary keys if there are no filters."""
    res = create_collection(
        "test",
        {
            "first": create_colspec("first", {"a": cs.Integer()}),
        },
    ).members()
    assert len(res) == 1
    assert not res["first"].is_optional
    assert issubclass(res["first"].col_spec, ColSpec)


def test_collection_no_common_primary_key():
    res = create_collection(
        "test",
        {
            "first": create_colspec("first", {"a": cs.Integer()}),
        },
        filters={"testfilter": Filter(lambda c: c.first.filter(pl.col("a") > 0))},
    ).members()
    assert len(res) == 1
    assert not res["first"].is_optional
    assert res["first"].col_spec.primary_keys() == []


def test_collection_primary_key_but_not_common():
    res = create_collection(
        "test",
        {
            "first": create_colspec("first", {"a": cs.Integer(primary_key=True)}),
            "second": create_colspec("second", {"b": cs.Integer(primary_key=True)}),
        },
        filters={"testfilter": Filter(lambda c: c.first.filter(pl.col("a") > 0))},
    ).members()
    assert len(res) == 2
    assert not res["first"].is_optional
    assert res["first"].col_spec.primary_keys() == ["a"]
    assert not res["second"].is_optional
    assert res["second"].col_spec.primary_keys() == ["b"]



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Sequence
from typing import Any, Generic, Protocol, Self, TypeVar

from ..optional_dependency import ColExpr
from ._base import Column

# ----------------------------------- ORDINAL MIXIN ---------------------------------- #


class Comparable(Protocol):
    def __gt__(self, other: Self, /) -> bool: ...
    def __ge__(self, other: Self, /) -> bool: ...


T = TypeVar("T", bound=Comparable)


class OrdinalMixin(Generic[T], Column):
    """Mixin to use for ordinal types."""

    def __init__(
        self,
        *,
        min: T | None = None,  # noqa: A002
        min_exclusive: T | None = None,
        max: T | None = None,  # noqa: A002
        max_exclusive: T | None = None,
        **kwargs: Any,
    ):
        if min is not None and min_exclusive is not None:
            raise ValueError("At most one of `min` and `min_exclusive` must be set.")
        if max is not None and max_exclusive is not None:
            raise ValueError("At most one of `max` and `max_exclusive` must be set.")

        if min is not None and max is not None and min > max:
            raise ValueError("`min` must not be greater than `max`.")
        if min_exclusive is not None and max is not None and min_exclusive >= max:
            raise ValueError("`min_exclusive` must not be greater or equal to `max`.")
        if min is not None and max_exclusive is not None and min >= max_exclusive:
            raise ValueError("`min` must not be greater or equal to `max_exclusive`.")
        if (
            min_exclusive is not None
            and max_exclusive is not None
            and min_exclusive >= max_exclusive
        ):
            raise ValueError(
                "`min_exclusive` must not be greater or equal to `max_exclusive`."
            )

        super().__init__(**kwargs)
        self.min = min
        self.min_exclusive = min_exclusive
        self.max = max
        self.max_exclusive = max_exclusive

    def validation_rules(self, expr: ColExpr) -> dict[str, ColExpr]:
        result = super().validation_rules(expr)
        if self.min is not None:
            result["min"] = expr >= self.min  # type: ignore
        if self.min_exclusive is not None:
            result["min_exclusive"] = expr > self.min_exclusive  # type: ignore
        if self.max is not None:
            result["max"] = expr <= self.max  # type: ignore
        if self.max_exclusive:
            result["max_exclusive"] = expr < self.max_exclusive  # type: ignore
        return result


# ------------------------------------ IS IN MIXIN ----------------------------------- #

U = TypeVar("U")


class IsInMixin(Generic[U], Column):
    """Mixin to use for types implementing "is in"."""

    def __init__(self, *, is_in: Sequence[U] | None = None, **kwargs: Any):
        super().__init__(**kwargs)
        self.is_in = is_in

    def validation_rules(self, expr: ColExpr) -> dict[str, ColExpr]:
        result = super().validation_rules(expr)
        if self.is_in is not None:
            result["is_in"] = expr.is_in(*self.is_in)
        return result



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
import pydiverse.colspec.collection
from pydiverse.colspec.optional_dependency import dy, pl

pytestmark = pytest.mark.skipif(
    dy.Column is None, reason="dataframely is required for this test"
)


class TestColSpec(cs.ColSpec):
    a = cs.Integer()


class MyCollection(pydiverse.colspec.collection.Collection):
    first: TestColSpec
    second: TestColSpec | None


def test_collection_optional_member():
    MyCollection.validate_polars_data({"first": pl.LazyFrame({"a": [1, 2, 3]})})


def test_filter_failure_info_keys_only_required():
    out, failure = MyCollection.filter_polars_data(
        {"first": pl.LazyFrame({"a": [1, 2, 3]})}
    )
    assert out.second is None
    assert set(failure.keys()) == {"first"}


def test_filter_failure_info_keys_required_and_optional():
    out, failure = MyCollection.filter_polars_data(
        {
            "first": pl.LazyFrame({"a": [1, 2, 3]}),
            "second": pl.LazyFrame({"a": [1, 2, 3]}),
        },
    )
    assert out.second is not None
    assert set(failure.keys()) == {"first", "second"}



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy
import decimal
import math
from collections.abc import Callable

import pydiverse.common as pdc

from ..optional_dependency import ColExpr, dy, sa
from ._base import Column
from ._mixins import OrdinalMixin


class Decimal(OrdinalMixin[decimal.Decimal], Column):
    """A column of decimal values with given precision and scale."""

    def __init__(
        self,
        precision: int | None = None,
        scale: int | None = None,
        *,
        nullable: bool = True,
        primary_key: bool = False,
        min: decimal.Decimal | float | int | None = None,  # noqa: A002
        min_exclusive: decimal.Decimal | float | int | None = None,
        max: decimal.Decimal | float | int | None = None,  # noqa: A002
        max_exclusive: decimal.Decimal | float | int | None = None,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
    ):
        """
        Args:
            precision: Maximum number of digits in each number.
            scale: Number of digits to the right of the decimal point in each number.
            nullable: Whether this column may contain null values.
            primary_key: Whether this column is part of the primary key of the schema.
                If ``True``, ``nullable`` is automatically set to ``False``.
            min: The minimum value for decimals in this column (inclusive).
            min_exclusive: Like ``min`` but exclusive. May not be specified if ``min``
                is specified and vice versa.
            max: The maximum value for decimals in this column (inclusive).
            max_exclusive: Like ``max`` but exclusive. May not be specified if ``max``
                is specified and vice versa.
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name.
        """
        if min is not None:
            _validate(min, precision, scale, "min")
            if isinstance(min, decimal.Decimal):
                min = float(min)  # noqa: A001, Decimal is more a RDBMS thing
        if min_exclusive is not None:
            _validate(min_exclusive, precision, scale, "min_exclusive")
            if isinstance(min_exclusive, decimal.Decimal):
                min_exclusive = float(min_exclusive)  # Decimal is more a RDBMS thing
        if max is not None:
            _validate(max, precision, scale, "max")
            if isinstance(max, decimal.Decimal):
                max = float(max)  # noqa: A001, Decimal is more a RDBMS thing
        if max_exclusive is not None:
            _validate(max_exclusive, precision, scale, "max_exclusive")
            if isinstance(max_exclusive, decimal.Decimal):
                max_exclusive = float(max_exclusive)  # Decimal is more a RDBMS thing

        super().__init__(
            nullable=nullable,
            primary_key=primary_key,
            min=min,
            min_exclusive=min_exclusive,
            max=max,
            max_exclusive=max_exclusive,
            check=check,
            alias=alias,
        )
        self.precision = precision
        self.scale = scale

    def dtype(self) -> pdc.Decimal:
        return pdc.Decimal()

    def to_dataframely(self) -> dy.Column:
        if self.scale is None or any(
            isinstance(x, float | int)
            for x in [self.min, self.max, self.min_exclusive, self.max_exclusive]
        ):
            ret = copy.copy(self)
            # in colspec we don't use python decimals for defining boundaries
            ret.min = decimal.Decimal(ret.min) if ret.min is not None else None
            ret.max = decimal.Decimal(ret.max) if ret.max is not None else None
            ret.min_exclusive = (
                decimal.Decimal(ret.min_exclusive)
                if ret.min_exclusive is not None
                else None
            )
            ret.max_exclusive = (
                decimal.Decimal(ret.max_exclusive)
                if ret.max_exclusive is not None
                else None
            )
            # polars cannot deal with scale=None as opposed to SQL
            ret.scale = ret.scale or 0
            return ret.to_dataframely()
        else:
            return super().to_dataframely()

    def sqlalchemy_column(self, name: str, dialect: sa.Dialect) -> sa.Column:
        """Obtain the SQL column specification of this column definition.

        Args:
            name: The name of the column.
            dialect: The SQL dialect for which to generate the column specification.

        Returns:
            The column as specified in :mod:`sqlalchemy`.
        """
        _ = dialect  # may be used in the future
        return sa.Column(
            name,
            sa.Numeric(
                self.precision or (38 if self.scale is not None else None), self.scale
            ),
            nullable=self.nullable,
            primary_key=self.primary_key,
            autoincrement=False,
        )


# --------------------------------------- UTILS -------------------------------------- #


def _validate(
    value: decimal.Decimal | int | float,
    precision: int | None,
    scale: int | None,
    name: str,
):
    if not isinstance(value, decimal.Decimal):
        value = decimal.Decimal(value)
    exponent = value.as_tuple().exponent
    if not isinstance(exponent, int):
        raise ValueError(f"Encountered 'inf' or 'NaN' for `{name}`.")
    if exponent is not None and scale is not None and -exponent > scale:
        raise ValueError(f"Scale of `{name}` exceeds scale of column.")
    if (
        precision is not None
        and scale is not None
        and _num_digits(int(value)) > precision - scale
    ):
        raise ValueError(f"`{name}` exceeds precision of column.")


def _num_digits(i: int) -> int:
    return int(math.log10(i) + 1)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import (
    C,
    Generator,
    assert_frame_equal,
    dy,
    np,
    pdt,
    pl,
)


class MySimpleColSpec(cs.ColSpec):
    a = cs.Int64()
    b = cs.String()


class PrimaryKeyColSpec(cs.ColSpec):
    a = cs.Int64(primary_key=True)
    b = cs.String()


class CheckColSpec(cs.ColSpec):
    a = cs.UInt64()
    b = cs.UInt64()

    @cs.rule_polars()
    @staticmethod
    def a_ge_b() -> pl.Expr:
        return pl.col("a") >= pl.col("b")


class ComplexColSpec(cs.ColSpec):
    a = cs.UInt8(primary_key=True)
    b = cs.UInt8(primary_key=True)

    @cs.rule_polars()
    @staticmethod
    def a_greater_b() -> pl.Expr:
        return pl.col("a") > pl.col("b")

    @cs.rule_polars(group_by=["a"])
    @staticmethod
    def minimum_two_per_a() -> pl.Expr:
        return pl.len() >= 2


class LimitedComplexColSpec(cs.ColSpec):
    a = cs.UInt8(primary_key=True)
    b = cs.UInt8(primary_key=True)

    @cs.rule_polars()
    @staticmethod
    def a_greater_b() -> pl.Expr:
        return pl.col("a") > pl.col("b")

    @cs.rule_polars(group_by=["a"])
    @staticmethod
    def minimum_two_per_a() -> pl.Expr:
        # We cannot generate more than 768 rows with this rule
        return pl.len() <= 3


# --------------------------------------- TESTS -------------------------------------- #


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("n", [0, 1000])
def test_sample_deterministic(n: int):
    with dy.Config(max_sampling_iterations=1):
        df = MySimpleColSpec.sample_polars(n)
        MySimpleColSpec.validate_polars(df)
        tbl = pdt.Table(df)
        MySimpleColSpec.validate(tbl)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("col_spec", [PrimaryKeyColSpec, CheckColSpec, ComplexColSpec])
@pytest.mark.parametrize("n", [0, 1000])
def test_sample_fuzzy(col_spec, n: int):
    df = col_spec.sample_polars(n, generator=Generator(seed=42))
    assert len(df) == n
    col_spec.validate_polars(df)
    tbl = pdt.Table(df)
    col_spec.validate(tbl)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_fuzzy_failure():
    with pytest.raises(ValueError):
        with dy.Config(max_sampling_iterations=5):
            ComplexColSpec.sample_polars(1000, generator=Generator(seed=42))


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("n", [1, 1000])
def test_sample_overrides(n: int):
    df = CheckColSpec.sample_polars(n, overrides={"b": range(n)})
    CheckColSpec.validate_polars(df)
    tbl = pdt.Table(df)
    CheckColSpec.validate(tbl)
    assert len(df) == n
    assert df.get_column("b").to_list() == list(range(n))


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_sample_overrides_with_removing_groups():
    generator = Generator()
    n = 333  # we cannot use something too large here or we'll never return
    overrides = np.random.randint(100, size=n)
    df = LimitedComplexColSpec.sample_polars(n, generator, overrides={"b": overrides})
    LimitedComplexColSpec.validate_polars(df)
    tbl = pdt.Table(df)
    LimitedComplexColSpec.validate(tbl)
    assert len(df) == n
    assert df.get_column("b").to_list() == list(overrides)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("n", [1, 1000])
def test_sample_overrides_allow_no_fuzzy(n: int):
    with dy.Config(max_sampling_iterations=1):
        df = CheckColSpec.sample_polars(n, overrides={"b": [0] * n})
        CheckColSpec.validate_polars(df)
        tbl = pdt.Table(df)
        CheckColSpec.validate(tbl)
        assert len(df) == n
        assert df.get_column("b").to_list() == [0] * n


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("n", [1, 1000])
def test_sample_overrides_full(n: int):
    df = CheckColSpec.sample_polars(n)
    df_override = CheckColSpec.sample_polars(n, overrides=df.to_dict())
    assert_frame_equal(df, df_override)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_overrides_invalid_column():
    with pytest.raises(ValueError, match=r"not in the schema"):
        MySimpleColSpec.sample_polars(overrides={"foo": []})


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_overrides_invalid_length():
    with pytest.raises(ValueError, match=r"`num_rows` is different"):
        MySimpleColSpec.sample_polars(overrides={"a": [1, 2]})



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec import ColSpec
from pydiverse.colspec.optional_dependency import dy


def perform_dataframely_operation():
    class C(ColSpec):
        a = cs.Integer(min=0)
        b = cs.String

    df = C.sample_polars(10)
    C.validate_polars(df)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_config_global():
    try:
        dy.Config.set_max_sampling_iterations(50)
        assert dy.Config.options["max_sampling_iterations"] == 50
        perform_dataframely_operation()
    finally:
        dy.Config.restore_defaults()


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_config_local():
    try:
        with dy.Config(max_sampling_iterations=35):
            assert dy.Config.options["max_sampling_iterations"] == 35
        assert dy.Config.options["max_sampling_iterations"] == 10_000
        perform_dataframely_operation()
    finally:
        dy.Config.restore_defaults()


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_config_local_nested():
    try:
        with dy.Config(max_sampling_iterations=35):
            assert dy.Config.options["max_sampling_iterations"] == 35
            with dy.Config(max_sampling_iterations=20):
                perform_dataframely_operation()
                assert dy.Config.options["max_sampling_iterations"] == 20
            assert dy.Config.options["max_sampling_iterations"] == 35
        assert dy.Config.options["max_sampling_iterations"] == 10_000
    finally:
        dy.Config.restore_defaults()


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_config_global_local():
    try:
        dy.Config.set_max_sampling_iterations(50)
        assert dy.Config.options["max_sampling_iterations"] == 50
        with dy.Config(max_sampling_iterations=35):
            perform_dataframely_operation()
            assert dy.Config.options["max_sampling_iterations"] == 35
        assert dy.Config.options["max_sampling_iterations"] == 50
    finally:
        dy.Config.restore_defaults()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections import defaultdict
from collections.abc import Callable

from .optional_dependency import ColExpr, pl

ValidationFunction = Callable[[], ColExpr] | staticmethod
ValidationFunctionPolars = Callable[[], pl.Expr] | staticmethod


class Rule:
    """Internal class representing validation rules."""

    def __init__(self, expr: ColExpr):
        self.expr = expr


class RulePolars:
    """Internal class representing validation rules."""

    def __init__(self, expr: pl.Expr):
        self.expr = expr

    @staticmethod
    def append_rules_polars(
        lf: pl.LazyFrame, rules: dict[str, "RulePolars"]
    ) -> pl.LazyFrame:
        return _with_evaluation_rules(lf, rules)


class GroupRule(Rule):
    """Rule that is evaluated on a group of columns."""

    def __init__(self, expr: ColExpr, group_columns: list[str]):
        super().__init__(expr)
        self.group_columns = group_columns


class GroupRulePolars(RulePolars):
    """Rule that is evaluated on a group of columns."""

    def __init__(self, expr: pl.Expr, group_columns: list[str]):
        super().__init__(expr)
        self.group_columns = group_columns


def rule(*, group_by: list[str] | None = None) -> Callable[[ValidationFunction], Rule]:
    """Mark a function as a rule to evaluate during validation.

    The name of the function will be used as the name of the rule. The function should
    return an expression providing a boolean value whether a row is valid wrt. the rule.
    A value of ``true`` indicates validity.

    Rules should be used only in the following two circumstances:

    - Validation requires accessing multiple columns (e.g. if valid values of column A
      depend on the value in column B).
    - Validation must be performed on groups of rows (e.g. if a column A must not
      contain any duplicate values among rows with the same value in column B).

    In all other instances, column-level validation rules should be preferred as it aids
    readability and improves error messages.

    Args:
        group_by: An optional list of columns to group by for rules operating on groups
            of rows. If this list is provided, the returned expression must return a
            single boolean value, i.e. some kind of aggregation function must be used
            (e.g. ``sum``, ``any``, ...).

    Note:
        You'll need to explicitly handle ``null`` values in your columns when defining
        rules. By default, any rule that evaluates to ``null`` because one of the
        columns used in the rule is ``null`` is interpreted as ``true``, i.e. the row
        is assumed to be valid.
    """

    def decorator(validation_fn: ValidationFunction) -> Rule:
        if isinstance(validation_fn, staticmethod):
            validation_fn = validation_fn.__wrapped__
        try:
            import pydiverse.transform  # noqa: F401
        except ImportError:
            # avoid running rule functions if pydiverse.transform is not installed
            return Rule(expr=ColExpr())
        if group_by is not None:
            return GroupRule(expr=validation_fn(), group_columns=group_by)
        return Rule(expr=validation_fn())

    return decorator


def rule_polars(
    *, group_by: list[str] | None = None
) -> Callable[[ValidationFunctionPolars], RulePolars]:
    """Mark a function as a rule to evaluate during validation.

    The name of the function will be used as the name of the rule. The function should
    return an expression providing a boolean value whether a row is valid wrt. the rule.
    A value of ``true`` indicates validity.

    Rules should be used only in the following two circumstances:

    - Validation requires accessing multiple columns (e.g. if valid values of column A
      depend on the value in column B).
    - Validation must be performed on groups of rows (e.g. if a column A must not
      contain any duplicate values among rows with the same value in column B).

    In all other instances, column-level validation rules should be preferred as it aids
    readability and improves error messages.

    Args:
        group_by: An optional list of columns to group by for rules operating on groups
            of rows. If this list is provided, the returned expression must return a
            single boolean value, i.e. some kind of aggregation function must be used
            (e.g. ``sum``, ``any``, ...).

    Note:
        You'll need to explicitly handle ``null`` values in your columns when defining
        rules. By default, any rule that evaluates to ``null`` because one of the
        columns used in the rule is ``null`` is interpreted as ``true``, i.e. the row
        is assumed to be valid.
    """

    def decorator(validation_fn: ValidationFunctionPolars) -> RulePolars:
        if pl.Expr is None:
            # decorator should also work if polars is not installed
            if group_by is not None:
                return GroupRulePolars(expr=None, group_columns=group_by)
            else:
                return RulePolars(expr=None)
        if isinstance(validation_fn, staticmethod):
            validation_fn = validation_fn.__wrapped__
        if group_by is not None:
            return GroupRulePolars(expr=validation_fn(), group_columns=group_by)
        return RulePolars(expr=validation_fn())

    return decorator


# ------------------------------------------------------------------------------------ #
#                                      EVALUATION                                      #
# ------------------------------------------------------------------------------------ #


def _with_evaluation_rules(
    lf: pl.LazyFrame, rules: dict[str, RulePolars]
) -> pl.LazyFrame:
    """Add evaluations of a set of rules on a data frame.

    Args:
        lf: The data frame on which to evaluate the rules.
        rules: The rules to evaluate where the key of the dictionary provides the name
            of the rule.

    Returns:
        The input lazy frame along with one boolean column for each rule with the name
        of the rule. For each rule, a value of ``True`` indicates successful validation
        while ``False`` indicates an issue.
    """
    # Rules must be distinguished into two types of rules:
    #  1. Simple rules can simply be selected on the data frame
    #  2. "Group" rules require a `group_by` and a subsequent join
    simple_exprs = {
        name: rule.expr
        for name, rule in rules.items()
        if not isinstance(rule, GroupRulePolars)
    }
    group_rules = {
        name: rule for name, rule in rules.items() if isinstance(rule, GroupRulePolars)
    }

    # Before we can select all of the simple expressions, we need to turn the
    # group rules into something to use in a `select` statement as well.
    return (
        # NOTE: A value of `null` always validates successfully as nullability should
        #  already be checked via dedicated rules.
        _with_group_rules(lf, group_rules).with_columns(
            **{name: expr.fill_null(True) for name, expr in simple_exprs.items()},
        )
    )


def _with_group_rules(
    lf: pl.LazyFrame, rules: dict[str, GroupRulePolars]
) -> pl.LazyFrame:
    # First, we partition the rules by group columns. This will minimize the number
    # of `group_by` calls and joins to make.
    grouped_rules: dict[frozenset[str], dict[str, pl.Expr]] = defaultdict(dict)
    for name, rule in rules.items():
        # NOTE: `null` indicates validity, see note above.
        grouped_rules[frozenset(rule.group_columns)][name] = rule.expr.fill_null(True)

    # Then, for each `group_by`, we apply the relevant rules and keep all the rule
    # evaluations around
    group_evaluations: dict[frozenset[str], pl.LazyFrame] = {}
    for group_columns, group_rules in grouped_rules.items():
        # We group by the group columns and apply all expressions
        group_evaluations[group_columns] = lf.group_by(group_columns).agg(**group_rules)

    # Eventually, we apply the rule evaluations onto the input data frame. For this,
    # we're using left-joins. This has two effects:
    #  - We're essentially "broadcasting" the results within each group across rows
    #    in the same group.
    #  - While an inner-join would be semantically more accurate, the left-join
    #    preserves the order of the left data frame.
    result = lf
    for group_columns, frame in group_evaluations.items():
        result = result.join(frame, on=list(group_columns), how="left")
    return result



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import inspect
import operator
import types
import typing
from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from functools import reduce
from pathlib import Path
from typing import Self

import structlog

from pydiverse.colspec.config import Config, alias_collection_fail, alias_subquery
from pydiverse.colspec.exc import (
    ColumnValidationError,
    ImplementationError,
    MemberValidationError,
    RuleValidationError,
    ValidationError,
)
from pydiverse.colspec.failure import FailureInfo
from pydiverse.colspec.optional_dependency import ColExpr, FrameType, dy, pdt, pl

from . import exc
from ._filter import Filter, FilterPolars
from .colspec import ColSpec, convert_to_dy_anno_dict


@dataclass
class MemberInfo:
    """Information about a member of a collection."""

    #: The schema of the member.
    col_spec: type[ColSpec]
    #: Whether the member is optional.
    is_optional: bool

    @staticmethod
    def is_member(anno: type):
        if dy.Column is not None:
            if inspect.isclass(anno) and issubclass(anno, dy.Schema):
                raise exc.AnnotationImplementationErrorDetail(
                    "Table annotations in Collections must not be a "
                    "dataframely Schema type. Use ColSpec instead.",
                    anno,
                )

        if isinstance(anno, types.UnionType):
            union_types = typing.get_args(anno)
            if dy.Column is not None:
                dy_schemas_in_union = sum(
                    1 if (inspect.isclass(t) and issubclass(t, dy.Schema)) else 0
                    for t in union_types
                )
                if dy_schemas_in_union > 0:
                    raise exc.AnnotationImplementationErrorDetail(
                        "Table annotations in Collections must not be a "
                        "dataframely Schema type. Use ColSpec instead. Found: "
                        f"{union_types}",
                        anno,
                    )
            col_specs_in_union = sum(
                1 if (inspect.isclass(t) and issubclass(t, ColSpec)) else 0
                for t in union_types
            )
            if col_specs_in_union > 1:
                raise exc.AnnotationImplementationErrorDetail(
                    "Table annotations in Collections must contain at most one "
                    f"ColSpec type. Found: {union_types}",
                    anno,
                )
            all_but_one_none = all(
                t is type(None)
                for t in union_types
                if not (inspect.isclass(t) and issubclass(t, ColSpec))
            )
            if not all_but_one_none and col_specs_in_union == 1:
                raise exc.AnnotationImplementationErrorDetail(
                    "Table annotations in Collections only allow None as second option "
                    f"next to ColSpec type. Found: {union_types}",
                    anno,
                )
            return (
                1 <= len(union_types) <= 2
                and col_specs_in_union == 1
                and all_but_one_none
            )
        return inspect.isclass(anno) and issubclass(anno, ColSpec)

    @staticmethod
    def new(anno: type):
        if isinstance(anno, types.UnionType):
            union_types = typing.get_args(anno)
            col_spec = [
                t for t in union_types if inspect.isclass(t) and issubclass(t, ColSpec)
            ][0]
            is_optional = any(t == type(None) for t in union_types)  # noqa: E721
        else:
            col_spec = anno
            is_optional = False
        return MemberInfo(col_spec, is_optional)

    @staticmethod
    def common_primary_keys(col_specs: Iterable[type[ColSpec]]) -> set[str]:
        return set.intersection(
            *[set(col_spec.primary_keys()) for col_spec in col_specs]
        )


class Collection:
    """Base class for all collections of tables with a predefined column specification.

    A collection is comprised of a set of *members* which are collectively "consistent",
    meaning they the collection ensures that invariants are held up *across* members.
    This is different to :mod:`dataframely` schemas which only ensure invariants
    *within* individual members.

    In order to properly ensure that invariants hold up across members, members must
    have a "common primary key", i.e. there must be an overlap of at least one primary
    key column across all members. Consequently, a collection is typically used to
    represent "semantic objects" which cannot be represented in a single table due
    to 1-N relationships that are managed in separate tables.

    A collection must only have type annotations for :class:`~pydiverse.colspec.ColSpec`
    with known column specification:

    .. code:: python
        class MyFirstColSpec:
            a: Integer

        class MyCollection(cs.Collection):
            first_member: MyFirstColSpec
            second_member: MySecondColSpec

    Besides, it may define *filters* (c.f. :meth:`~dataframely.filter`) and arbitrary
    methods.

    A colspec.Collection can also be instantiated and filled with
    pydiverse transform Table, pipedag Table objects, or pipedag task outputs which
    reference a table. This yields quite intuitive syntax:

    .. code:: python

        c = MyCollection.build()
        c.first_member = pipdag_task1()
        c.second_member = pipdag_task2()
        pipdag_task3(c)

    Attention:
        Do NOT use this class in combination with ``from __future__ import annotations``
        as it requires the proper schema definitions to ensure that the collection is
        implemented correctly.
    """

    def get_pdt(self, name: str, cfg: Config = Config.default) -> pdt.Table:
        tbl = getattr(self, name)
        if not isinstance(tbl, pdt.Table):
            raise TypeError(
                f"Collection member '{name}' is not a pydiverse transform Table, but "
                f"{type(tbl)}; collection={self}"
            )
        if cfg.fix_table_names:
            # fix the name of the table according to Collection member name
            tbl._ast.name = name
        return tbl

    def validate(self, *, cast: bool = False):
        out, failure = self.filter(cast=cast)
        if any(len(getattr(failure, tbl_name)) > 0 for tbl_name in failure.members()):
            raise MemberValidationError(
                {
                    tbl_name: RuleValidationError(getattr(failure, tbl_name).counts())
                    for tbl_name in failure.members()
                }
            )
        return out

    def validate_polars(self, *, cast: bool = False, fault_tolerant: bool = False):
        self.finalize()
        return self.validate_polars_data(
            self.__dict__, cast=cast, fault_tolerant=fault_tolerant
        )

    @classmethod
    def validate_polars_data(
        cls,
        data: Mapping[str, FrameType],
        *,
        cast: bool = False,
        fault_tolerant: bool = False,
    ) -> Self:
        """Validate that a set of data frames satisfy the collection's invariants.

        Args:
            data: The members of the collection which ought to be validated. The
                dictionary must contain exactly one entry per member with the name of
                the member as key.
            cast: Whether columns with a wrong data type in the member data frame are
                cast to their schemas' defined data types if possible.

        Raises:
            ValueError: If an insufficient set of input data frames is provided, i.e. if
                any required member of this collection is missing in the input.
            ValidationError: If any of the input data frames does not satisfy its schema
                definition or the filters on this collection result in the removal of at
                least one row across any of the input data frames.

        Returns:
            An instance of the collection. All members of the collection are guaranteed
            to be valid with respect to their respective schemas and the filters on this
            collection did not remove rows from any member.
        """
        import dataframely.exc as dy_exc
        import polars.exceptions as plexc

        DynCollection = convert_collection_to_dy(cls)
        logger_name = __name__ + "." + cls.__name__ + ".validate_polars"
        try:
            return cls.from_dy_collection(DynCollection.validate(data, cast=True))
        except dy_exc.ImplementationError as e:
            logger = structlog.getLogger(logger_name)
            logger.exception(
                "Dataframely raised column specification implementation error"
            )
            if not fault_tolerant:
                raise exc.ImplementationError(str(e))  # noqa: B904
        except plexc.InvalidOperationError as e:
            logger = structlog.getLogger(logger_name)
            logger.exception("Dataframely validation failed within polars expression")
            if not fault_tolerant:
                raise ValidationError(str(e))  # noqa: B904
        except dy_exc.ValidationError as e:
            logger = structlog.getLogger(logger_name)
            logger.exception("Dataframely validation failed")
            if not fault_tolerant:
                # Try to replicate exact error class. However, the constructor
                # does not always store the arguments to it directly.
                exc_class = getattr(exc, e.__class__.__name__)
                if hasattr(e, "errors"):
                    raise exc_class(e.errors)  # noqa: B904
                elif hasattr(e, "schema_errors") and hasattr(e, "column_errors"):
                    new_e = exc_class({})
                    new_e.schema_errors = e.schema_errors
                    new_e.column_errors = e.column_errors
                    raise new_e  # noqa: B904
                else:
                    raise ValidationError(str(e))  # noqa: B904
        return cls._init_polars_data(data)  # ignore validation if fault_tolerant

    def is_valid(
        self,
        *,
        cast: bool = False,
    ) -> bool:
        """Utility method to check whether :meth:`validate` raises an exception.

        Args:
            cast: Whether columns with a wrong data type in the member data frame are
                cast to their schemas' defined data types if possible.

        Returns:
            Whether the provided members satisfy the invariants of the collection.

        Raises:
            ValueError: If an insufficient set of input data frames is provided, i.e. if
                any required member of this collection is missing in the input.
        """
        try:
            self.validate(cast=cast)
            return True
        except ValidationError:
            return False

    def is_valid_polars(self, *, cast: bool = False, fault_tolerant: bool = False):
        self.finalize()
        return self.is_valid_polars_data(
            self.__dict__, cast=cast, fault_tolerant=fault_tolerant
        )

    @classmethod
    def is_valid_polars_data(
        cls,
        data: Mapping[str, FrameType],
        *,
        cast: bool = False,
        fault_tolerant: bool = False,
    ) -> bool:
        """Utility method to check whether :meth:`validate` raises an exception.

        Args:
            data: The members of the collection which ought to be validated. The
                dictionary must contain exactly one entry per member with the name of
                the member as key. The existence of all keys is checked via the
                :mod:`dataframely` mypy plugin.
            cast: Whether columns with a wrong data type in the member data frame are
                cast to their schemas' defined data types if possible.

        Returns:
            Whether the provided members satisfy the invariants of the collection.

        Raises:
            ValueError: If an insufficient set of input data frames is provided, i.e. if
                any required member of this collection is missing in the input.
        """
        try:
            cls.validate_polars_data(data, cast=cast, fault_tolerant=fault_tolerant)
            return True
        except ValidationError:
            return False

    def filter(
        self, *, cast: bool = False, cfg: Config = Config.default
    ) -> tuple[Self, Self]:
        """Filter rows which conform to column specifications and collections rules.

        Returns a tuple of two new collections one with the filtered tables as member
        variables and one with FailureInfo objects as member variables.
        """
        from pydiverse.transform.extended import (
            C,
            drop,
            full_join,
            group_by,
            left_join,
            mutate,
            select,
            summarize,
        )

        self.finalize(assert_pdt=True)

        members: dict[str, MemberInfo] = self.members()

        # first filter the tables individually, else invalid rows could cause incorrect
        # results in the multi-table filters
        individually_filtered = self.__class__.build()
        table_level_fail = self.__class__.build()

        for name, member in members.items():
            out, failure = member.col_spec.filter(self.get_pdt(name, cfg), cast=cast)
            setattr(individually_filtered, name, out)
            setattr(table_level_fail, name, failure)

        # join tables needed for executing filter rules
        join_members: dict[str, set[str]] = {name: set() for name in members}
        extra_rules: dict[str, dict[str, ColExpr]] = {name: {} for name in members}

        # for reuse of subqueries across different tables:
        @dataclass
        class GroupSubquery:
            # grouping keys of subquery
            keys: set[str]
            # table names that need subquery for filtering
            tbls: set[str]
            # table names joined in subquery
            join_tbls: set[str]
            # dict[filter name, filter expression]
            cols: dict[str, ColExpr]

            @staticmethod
            def build():
                return GroupSubquery(set(), set(), set(), {})

        group_subqueries: dict[tuple[str], GroupSubquery] = {}

        for pred_name, pred in self.filter_rules().items():
            logic = pred.logic_fn(self)
            expr_tbl_names = [
                tbl_name
                for tbl_name in members.keys()
                if logic.uses_table(self.get_pdt(tbl_name, cfg))
            ]
            expr_col_specs = [members[tbl_name].col_spec for tbl_name in expr_tbl_names]

            expr_pk_union = self._pk_union(*expr_col_specs)

            for name in self.members().keys():
                tbl = members[name].col_spec
                join = self.get_join(name, set(expr_tbl_names) - {name}, cfg=cfg)
                pk_overlap = expr_pk_union.intersection(tbl.primary_keys())
                requires_grouping = (
                    len(expr_pk_union.difference(tbl.primary_keys())) > 0
                )

                if join is not None:
                    if requires_grouping:
                        key = tuple(
                            *pk_overlap, 0, *sorted(tbl.primary_keys() + expr_pk_union)
                        )
                        if key not in group_subqueries:
                            group_subqueries[key] = GroupSubquery.build()
                        group_subqueries[key].keys = pk_overlap
                        group_subqueries[key].tbls |= name
                        group_subqueries[key].join_tbls |= set(expr_tbl_names)
                        group_subqueries[key].cols[pred_name] = logic
                    else:
                        join_members[name] |= set(expr_tbl_names) - {name}
                        extra_rules[name][pred.logic_fn.__name__] = logic

        join_subqueries: dict[str, list[tuple[pdt.Table, set[str]]]] = {
            name: [] for name in members
        }

        for group_subquery in group_subqueries.values():
            for name in group_subquery.tbls:
                subquery = (
                    self._get_join(*group_subquery.join_tbls)
                    >> group_by(group_subquery.keys)
                    >> summarize(**group_subquery.cols)
                )
                join_subqueries[name].append((subquery, group_subquery.keys))
                for col in group_subquery.cols.keys():
                    extra_rules[name][col] = subquery[col]

        new = self.__class__.build()
        fail = self.__class__.build()

        for name in self.members().keys():
            tbl = getattr(individually_filtered, name)
            col_spec = self.member_col_specs()[name]

            if len(join_members[name]) > 0:
                filter_join = individually_filtered.get_join(
                    name, *join_members[name], cfg=cfg
                )
            else:
                filter_join = tbl
            if len(join_subqueries[name]):
                for subquery, keys in join_subqueries[name]:
                    filter_join >>= left_join(subquery >> alias_subquery(cfg), on=keys)

            # NOTE: we currently throw rows out if a rule results in null. We could also
            # keep everything, but then we should do the same if no match in the join
            # is found.
            extra_rules[name] = {
                name: rule.fill_null(False) for name, rule in extra_rules[name].items()
            }

            # The above left_join should not duplicate rows of `tbl`, so we don't need
            # to care about uniqueness here.
            ok_rows = (
                filter_join
                >> pdt.filter(*extra_rules.get(name).values())
                >> select(*tbl)
            )
            setattr(new, name, ok_rows)

            collection_level_invalid_rows = (
                filter_join
                >> pdt.filter(~pdt.all(True, *extra_rules.get(name).values()))
                >> mutate(**extra_rules.get(name))
            )
            table_level_invalid_rows: pdt.Table = getattr(
                table_level_fail, name
            )._invalid_rows

            rule_columns: dict[str, ColExpr] = (
                getattr(table_level_fail, name).rule_columns | extra_rules[name]
            )

            failure = (
                table_level_invalid_rows
                >> full_join(
                    (
                        coll_failure := collection_level_invalid_rows
                        >> alias_collection_fail(
                            cfg, self.__class__.__name__.lower() + "_coll_invalid"
                        )
                    )
                    >> drop(*col_spec.column_names()),
                    on=[
                        # collection_level_invalid_rows is based on
                        # individually_filtered and thus has unique primary keys
                        table_level_invalid_rows[pk] == coll_failure[pk]
                        for pk in col_spec.primary_keys()
                    ],
                )
                >> mutate(
                    **{
                        col_name: table_level_invalid_rows[col_name].fill_null(
                            coll_failure[col_name]
                        )
                        for col_name in col_spec.column_names()
                    }
                )
                >> mutate(
                    **{rule: C[rule].fill_null(True) for rule in rule_columns.keys()}
                )
            )

            original_tbl = getattr(table_level_fail, name).tbl
            setattr(fail, name, FailureInfo(original_tbl, failure, rule_columns, cfg))

        return new, fail

    def filter_rules(self) -> dict[str, Filter]:
        rules = {
            pred: getattr(self, pred)
            for pred in dir(self)
            if isinstance(getattr(self, pred), Filter)
        }
        if "_primary_key_" in rules:
            raise ImplementationError(
                "Collection cannot have a filter named '_primary_key_'"
            )
        return rules

    def _pk_overlap(
        self, tbl: str | type[ColSpec], *more_tbls: str | type[ColSpec]
    ) -> set[str]:
        tbls: list[ColSpec] = [
            self.member_col_specs()[t] if isinstance(t, str) else t
            for t in (tbl, *more_tbls)
        ]
        return set(tbls[0].primary_keys()).intersection(
            *(other.primary_keys() for other in tbls[1:])
        )

    def _pk_union(
        self, tbl: str | type[ColSpec], *more_tbls: Iterable[str | type[ColSpec]]
    ) -> set[str]:
        tbls: list[ColSpec] = [
            self.member_col_specs()[t] if isinstance(t, str) else t
            for t in (tbl, *more_tbls)
        ]
        return set(tbls[0].primary_keys()).union(
            *(other.primary_keys() for other in tbls[1:])
        )

    def _get_join(
        self, *tbls: Iterable[str], cfg: Config = Config.default
    ) -> pdt.Table | None:
        """
        Similar to get_join(), but without given leftmost table.

        It is used for constructing grouped subqueries.
        """
        col_specs = self.member_col_specs()
        primary_keyss = {
            name: spec.primary_keys()
            for name, spec in col_specs.items()
            if name in tbls
        }
        # the ordering should match that of get_join()
        ordered_tbls = sorted(
            primary_keyss.keys(), key=lambda name: (len(primary_keyss[name]), name)
        )
        return self.get_join(*ordered_tbls, cfg=cfg)

    def get_join(
        self, tbl: str, *more_tbls: Iterable[str], cfg: Config = Config.default
    ) -> pdt.Table | None:
        """
        Get a left join expression if tables should be joinable.

        This method is intended to be overridden in case the automatic detection by
        checking primary key overlap is not sufficient. The automatic detection assumes
        that when ordering more_tbls by number of primary keys, that every next table
        can be joined with its primary key columns to the left most table that has them.

        This heuristic fails for some cases where not all tables have at least one
        common primary key column. In such situations it is intended to be overridden
        with a version that solves the problem for the concrete tables in a collection.
        """
        more_tbls = list(more_tbls)
        result = self.get_pdt(tbl, cfg)
        col_specs = self.member_col_specs()
        primary_keyss = {
            name: spec.primary_keys()
            for name, spec in col_specs.items()
            if name in more_tbls
        }
        # it is important to ensure consistent ordering in case additional tables are
        # added to `more_tbls`
        ordered_tbls = sorted(
            primary_keyss.keys(), key=lambda name: (len(primary_keyss[name]), name)
        )
        primary_keyss[tbl] = col_specs[tbl].primary_keys()
        ordered_tbls = [tbl, *ordered_tbls]
        for i, name in enumerate(ordered_tbls):
            if i > 0:
                join_tbls = ordered_tbls[0:i]
                pk_set = set(primary_keyss[name])
                on = None
                for join_name in join_tbls:
                    pk_overlap = pk_set.intersection(primary_keyss[join_name])
                    if len(pk_overlap) > 0:
                        on = reduce(
                            operator.and_,
                            (
                                self.get_pdt(name, cfg)[f]
                                == self.get_pdt(join_name, cfg)[f]
                                for f in pk_overlap
                            ),
                            on or pdt.lit(True),
                        )
                        pk_set -= pk_overlap
                if on is None:
                    # one join table has no matching primary keys to previous tables
                    logger = structlog.getLogger(
                        __name__,
                        collection=self,
                        method="get_join",
                        members=self.members(),
                    )
                    logger.debug(
                        "Heuristic failed to join table {name}. "
                        "Please override get_join method in this collection.",
                        join_tbls=join_tbls,
                        primary_key=pk_set,
                    )
                    return None
                result = result >> pdt.left_join(self.get_pdt(name, cfg), on)
        return result

    def filter_polars(
        self, *, cast: bool = False
    ) -> tuple[Self, dict[str, dy.FailureInfo]]:
        self.finalize()
        return self.filter_polars_data(self.__dict__, cast=cast)

    @classmethod
    def filter_polars_data(
        cls, data: Mapping[str, FrameType], *, cast: bool = False
    ) -> tuple[Self, dict[str, dy.FailureInfo]]:
        DynCollection = convert_collection_to_dy(cls)
        coll, failure = DynCollection.filter(data, cast=cast)
        return cls.from_dy_collection(coll), failure

    def cast_polars(self) -> Self:
        self.finalize()
        return self.cast_polars_data(self.__dict__)

    @classmethod
    def cast_polars_data(cls, data: Mapping[str, FrameType]) -> Self:
        DynCollection = convert_collection_to_dy(cls)
        try:
            return cls.from_dy_collection(DynCollection.cast(data))
        except pl.exceptions.ColumnNotFoundError as e:
            # TODO: improve error message by checking missing and extra columns for
            #  all member tables
            raise ColumnValidationError() from e

    # -------------------------------- Member inquiries --------------------------- #

    @classmethod
    def members(cls) -> dict[str, MemberInfo]:
        """Information about the members of the collection."""

        def better_msg(fn, arg, *, name: str):
            try:
                return fn(arg)
            except exc.AnnotationImplementationErrorDetail as e:
                raise exc.AnnotationImplementationError(name, e._type) from e

        return {
            k: MemberInfo.new(v)
            for k, v in typing.get_type_hints(cls).items()
            if better_msg(MemberInfo.is_member, v, name=k)
        }

    @classmethod
    def member_col_specs(cls) -> dict[str, type[ColSpec]]:
        """The column specifications of all members of the collection."""
        return {
            k: MemberInfo.new(v).col_spec
            for k, v in typing.get_type_hints(cls).items()
            if MemberInfo.is_member(v)
        }

    @classmethod
    def required_members(cls) -> set[str]:
        """The names of all required members of the collection."""
        return {
            k
            for k, v in typing.get_type_hints(cls).items()
            if MemberInfo.is_member(v) and not MemberInfo.new(v).is_optional
        }

    @classmethod
    def optional_members(cls) -> set[str]:
        """The names of all optional members of the collection."""
        return {
            k
            for k, v in typing.get_type_hints(cls).items()
            if MemberInfo.is_member(v) and MemberInfo.new(v).is_optional
        }

    @classmethod
    def common_primary_keys(cls) -> list[str]:
        """The primary keys which are shared by all members of the collection."""
        return sorted(MemberInfo.common_primary_keys(cls.member_col_specs().values()))

    def to_dict(self) -> dict[str, ColSpec]:
        """Return a dictionary representation of this collection."""
        return {
            member: getattr(self, member)
            for member in self.member_col_specs()
            if getattr(self, member) is not None
        }

    # ---------------------------------- COLLECTION ---------------------------------- #

    def collect_all_polars(self) -> Self:
        """Collect all members of the collection.

        This method collects all members in parallel for maximum efficiency. It is
        particularly useful when :meth:`filter` is called with lazy frame inputs.

        Returns:
            The same collection with all members collected once.

        Note:
            As all collection members are required to be lazy frames, the returned
            collection's members are still "lazy". However, they are "shallow-lazy",
            meaning they are obtained by calling ``.collect().lazy()``.
        """
        import polars.exceptions as plexc

        try:
            dfs = pl.collect_all([lf for lf in self.to_dict().values()])
        except plexc.PolarsError as e:
            raise ValidationError(str(e)) from e
        return self._init_polars_data(
            {key: dfs[i].lazy() for i, key in enumerate(self.to_dict().keys())}
        )

    # -------------------------- Polars/Parquet PERSISTENCE ----------------------- #

    def write_parquet(self, directory: Path):
        self.finalize()
        DynCollection = convert_collection_to_dy(self.__class__)
        coll = DynCollection._init(
            {k: v for k, v in self.__dict__.items() if v is not None}
        )
        coll.write_parquet(directory)

    @classmethod
    def read_parquet(cls, directory: Path) -> Self:
        DynCollection = convert_collection_to_dy(cls)
        return DynCollection.read_parquet(directory)

    @classmethod
    def scan_parquet(cls, directory: Path) -> Self:
        DynCollection = convert_collection_to_dy(cls)
        return DynCollection.scan_parquet(directory)

    # ---------------------------------- BUILDING --------------------------------- #

    @classmethod
    def build(cls):
        try:
            return cls(**{member: None for member in cls.__annotations__.keys()})
        except TypeError:
            try:
                return cls(**{member: None for member in cls.members().keys()})
            except TypeError:
                try:
                    return cls()
                except TypeError:
                    raise ImplementationError(  # noqa: B904
                        "Failed constructing collection with empty members. Try adding"
                        " @dataclasses.dataclass annotation to a collection class you"
                        " like to build."
                    )

    def finalize(self, assert_pdt=False):
        # finalize builder stage and ensure that all dataclass members have been set
        errors = {
            member
            for member, info in self.members().items()
            if not info.is_optional and getattr(self, member) is None
        }
        assert len(errors) == 0, (
            f"Dataclass building was not finalized before usage. "
            f"Please make sure to assign the following members "
            f"on '{self}': {','.join(errors)}"
        )
        if assert_pdt:
            errors = {
                member: type(getattr(self, member))
                for member, info in self.members().items()
                if getattr(self, member) is not None
                and not isinstance(getattr(self, member), pdt.Table)
            }
            assert len(errors) == 0, (
                f"Collection includes other member type than expected. "
                f"The function you called expects pdt.Table members "
                f"in '{self}': {','.join(errors)}"
            )

    # ----------------------------------- UTILITIES ---------------------------------- #

    @classmethod
    def _init_polars_data(cls, data: Mapping[str, FrameType]) -> Self:
        out = cls.build()
        for member_name, member in cls.members().items():
            if member.is_optional and (
                member_name not in data or data[member_name] is None
            ):
                setattr(out, member_name, None)
            else:
                setattr(out, member_name, data[member_name].lazy())
        return out

    @classmethod
    def from_dy_collection(cls, c: dy.Collection) -> Self:
        return cls._init_polars_data(
            {name: getattr(c, name) for name in c.members().keys()}
        )

    @classmethod
    def _validate_polars_input_keys(cls, data: Mapping[str, FrameType]):
        actual = set(data)

        missing = cls.required_members() - actual
        if len(missing) > 0:
            raise ValueError(
                f"Input misses {len(missing)} required members: {', '.join(missing)}."
            )

        superfluous = actual - set(cls.members())
        if len(superfluous) > 0:
            logger = structlog.getLogger(__name__ + "." + cls.__name__ + ".cast")
            logger.warning(
                f"Input provides {len(superfluous)} superfluous members that are "
                f"ignored: {', '.join(superfluous)}."
            )

    def pk_is_null(self, tbl: pdt.Table) -> ColExpr:
        tbl_name = next(attr for attr in dir(self) if getattr(self, attr) == tbl)
        return tbl[self.member_col_specs()[tbl_name].primary_keys()[0]].is_null()


def convert_filter_to_dy(f: FilterPolars):
    return dy._filter.Filter(f.logic)


def convert_collection_to_dy(
    collection: Collection | type[Collection],
) -> type[dy.Collection]:
    from pydiverse.colspec import FilterPolars

    cls = collection.__class__ if isinstance(collection, Collection) else collection
    filters = {
        k: convert_filter_to_dy(v)
        for k, v in collection.__dict__.items()
        if isinstance(v, FilterPolars)
    }
    DynCollection = type[dy.Collection](
        cls.__name__,
        (dy.Collection,),
        {
            "__annotations__": convert_to_dy_anno_dict(typing.get_type_hints(cls)),
            **filters,
        },
    )  # type:type[dy.Collection]
    return DynCollection



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec._rule import GroupRulePolars, RulePolars
from pydiverse.colspec.exc import ImplementationError, RuleImplementationError
from pydiverse.colspec.optional_dependency import dy, pl
from pydiverse.colspec.testing.factory import create_colspec


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_group_rule_group_by_error():
    with pytest.raises(
        ImplementationError,
        match=(
            r"Group validation rule 'b_greater_zero' has been implemented "
            r"incorrectly\. It references 1 columns which are not in the schema"
        ),
    ):
        create_colspec(
            "test",
            columns={"a": cs.Integer(), "b": cs.Integer()},
            rules={
                "b_greater_zero": GroupRulePolars(
                    (pl.col("b") > 0).all(), group_columns=["c"]
                )
            },
        ).validate_polars(None)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_rule_implementation_error():
    with pytest.raises(
        RuleImplementationError, match=r"rule 'integer_rule'.*returns dtype 'Int64'"
    ):
        create_colspec(
            "test",
            columns={"a": cs.Integer()},
            rules={"integer_rule": RulePolars(pl.col("a") + 1)},
        ).validate_polars(None)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_group_rule_implementation_error():
    with pytest.raises(
        RuleImplementationError,
        match=(
            r"rule 'b_greater_zero'.*returns dtype 'List\(Boolean\)'.*"
            r"make sure to use an aggregation function"
        ),
    ):
        create_colspec(
            "test",
            columns={"a": cs.Integer(), "b": cs.Integer()},
            rules={
                "b_greater_zero": GroupRulePolars(pl.col("b") > 0, group_columns=["a"])
            },
        ).validate_polars(None)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_rule_column_overlap_error():
    with pytest.raises(
        ImplementationError,
        match=r"Rules and columns must not be named equally but found 1 overlaps",
    ):
        create_colspec(
            "test",
            columns={"test": cs.Integer(alias="a")},
            rules={"a": RulePolars(pl.col("a") > 0)},
        ).validate_polars(None)



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Callable, Sequence

import pydiverse.common as pdc

from ..optional_dependency import ColExpr, sa
from ._base import Column
from .string import String


class Enum(Column):
    """A column of enum (string) values."""

    def __init__(
        self,
        categories: Sequence[str],
        *,
        nullable: bool = True,
        primary_key: bool = False,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
    ):
        """
        Args:
            categories: The list of valid categories for the enum.
            nullable: Whether this column may contain null values.
            primary_key: Whether this column is part of the primary key of the schema.
                If ``True``, ``nullable`` is automatically set to ``False``.
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name.
        """
        super().__init__(
            nullable=nullable, primary_key=primary_key, check=check, alias=alias
        )
        self.categories = categories

    def dtype(self) -> pdc.Dtype:
        raise NotImplementedError(
            "Enum column type is not yet implemented in pydiverse libraries."
        )

    def sqlalchemy_column(self, name: str, dialect: sa.Dialect) -> sa.Column:
        """Obtain the SQL column specification of this column definition.

        Args:
            name: The name of the column.
            dialect: The SQL dialect for which to generate the column specification.

        Returns:
            The column as specified in :mod:`sqlalchemy`.
        """
        _ = dialect  # may be used in the future
        str_type = String(
            nullable=self.nullable,
            primary_key=self.primary_key,
            min_length=min(len(cat) for cat in self.categories),
            max_length=max(len(cat) for cat in self.categories),
        )
        return str_type.sqlalchemy_column(name, dialect)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.colspec import convert_to_dy_col_spec
from pydiverse.colspec.exc import ImplementationError
from pydiverse.colspec.optional_dependency import dy


class MyFirstColSpec(cs.ColSpec):
    a = cs.UInt8(primary_key=True)


class MySecondColSpec(cs.ColSpec):
    a = cs.UInt16(primary_key=True)
    b = cs.Integer


class MyThirdColSpec(MyFirstColSpec, MySecondColSpec):
    pass


class MyFourthColSpec(MyFirstColSpec, MySecondColSpec):
    c = cs.Float64()


def test_columns():
    assert sorted(MyFirstColSpec.column_names()) == ["a"]
    assert sorted(MySecondColSpec.column_names()) == ["a", "b"]
    assert sorted(MyThirdColSpec.column_names()) == ["a", "b"]
    assert sorted(MyFourthColSpec.column_names()) == ["a", "b", "c"]


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_columns_dataframely():
    first = convert_to_dy_col_spec(MyFirstColSpec)
    second = convert_to_dy_col_spec(MySecondColSpec)
    third = convert_to_dy_col_spec(MyThirdColSpec)
    fourth = convert_to_dy_col_spec(MyFourthColSpec)
    assert sorted(first.column_names()) == ["a"]
    assert sorted(second.column_names()) == ["a", "b"]
    assert sorted(third.column_names()) == ["a", "b"]
    assert sorted(fourth.column_names()) == ["a", "b", "c"]


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_dataframely_columns_fail():
    class FailColSpec(cs.ColSpec):
        a = cs.Float64()
        b = dy.String()

    with pytest.raises(
        ImplementationError, match="Dataframely Columns won't work in ColSpec classes."
    ):
        FailColSpec.column_names()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Iterable
from typing import Literal

from pydiverse.colspec.exc import (
    ColumnValidationError,
    DtypeValidationError,
)
from pydiverse.common import Dtype

from .columns import Column
from .optional_dependency import pdt

DtypeCasting = Literal["none", "lenient", "strict"]


def validate_columns(
    tbl: pdt.Table,
    expected: Iterable[str],
) -> pdt.Table:
    """Validate the existence of expected columns in a table.

    Args:
        tbl: The table whose list of columns to validate.
        expected: The list of columns that _should_ be observed.

    Raises:
        ValidationError: If any expected column is not part of the actual columns.

    Returns:
        The input table, either as-is or with extra columns stripped.
    """
    actual_set = set(col.name for col in tbl)
    expected_set = set(expected)

    missing_columns = expected_set - actual_set
    if len(missing_columns) > 0:
        raise ColumnValidationError(
            list(sorted(missing_columns)),
            extra=list(sorted(actual_set - expected_set)),
            actual=list(sorted(actual_set)),
        )

    return tbl >> pdt.select(*expected)


def validate_dtypes(
    tbl: pdt.Table,
    expected: dict[str, Column],
    *,
    casting: Literal["none", "lenient", "strict"],
) -> pdt.Table:
    """Validate the dtypes of all expected columns in a table.

    Args:
        tbl: The table whose column dtypes to validate.
        expected: The column definitions carrying the expected dtypes.
        casting: The strategy for casting dtypes.

    Raises:
        DtypeValidationError: If the expected column dtypes do not match the input's and
            ``casting`` set to ``none``.

    Returns:
        The input table with all column dtypes ensured to have the expected dtype.
    """
    from pydiverse.transform import C, mutate

    dtype_errors: dict[str, tuple[Dtype, Dtype]] = {}
    for name, col in expected.items():
        if not tbl[name].dtype().is_subtype(col.dtype()):
            dtype_errors[name] = (tbl[name].dtype(), col.dtype())

    if len(dtype_errors) > 0:
        if casting == "none":
            raise DtypeValidationError(dtype_errors)
        else:
            return tbl >> mutate(
                **{
                    name: C[name].cast(
                        expected[name].dtype(), strict=(casting == "strict")
                    )
                    for name in dtype_errors.keys()
                }
            )

    return tbl



# Copyright (c) QuantCo and pydiverse contributors 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import assert_frame_equal, dy, pl
from pydiverse.colspec.testing import evaluate_rules_polars, rules_from_exprs_polars


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_validate_min_length():
    column = cs.String(min_length=2)
    lf = pl.LazyFrame({"a": ["foo", "x"]})
    actual = evaluate_rules_polars(
        lf, rules_from_exprs_polars(column.validation_rules(pl.col("a")))
    )
    expected = pl.LazyFrame({"min_length": [True, False]})
    assert_frame_equal(actual, expected)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_validate_max_length():
    column = cs.String(max_length=2)
    lf = pl.LazyFrame({"a": ["foo", "x"]})
    actual = evaluate_rules_polars(
        lf, rules_from_exprs_polars(column.validation_rules(pl.col("a")))
    )
    expected = pl.LazyFrame({"max_length": [False, True]})
    assert_frame_equal(actual, expected)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_validate_regex():
    column = cs.String(regex="[0-9][a-z]$")
    lf = pl.LazyFrame({"a": ["33x", "3x", "44"]})
    actual = evaluate_rules_polars(
        lf, rules_from_exprs_polars(column.validation_rules(pl.col("a")))
    )
    expected = pl.LazyFrame({"regex": [True, True, False]})
    assert_frame_equal(actual, expected)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_validate_all_rules():
    column = cs.String(nullable=False, min_length=2, max_length=4)
    lf = pl.LazyFrame({"a": ["foo", "x", "foobar", None]})
    actual = evaluate_rules_polars(
        lf, rules_from_exprs_polars(column.validation_rules(pl.col("a")))
    )
    expected = pl.LazyFrame(
        {
            "min_length": [True, False, True, True],
            "max_length": [True, True, False, True],
            "nullability": [True, True, True, False],
        }
    )
    assert_frame_equal(actual, expected, check_column_order=False)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Callable
from dataclasses import dataclass

from .columns._utils import classproperty
from .optional_dependency import pdt, verb


@dataclass
class Config:
    """
    Configuration for ColSpec operations.

    ColSpec and Collection classes do not handle configuration directly. Instead,
    you must pass a Config instance explicitly to validation or filtering methods.
    Config also enables integration with other libraries in the pydiverse
    ecosystem—for example, by supplying a materialization hook that prevents ColSpec
    from generating subqueries, which can help avoid performance bottlenecks.
    """

    #: If fix_table_names=True, table names are adjusted to match the corresponding
    # ColSpec class name.
    fix_table_names: bool = True
    #: If stop_validation_on_column_error=True, validation stops in case of column
    # type errors before looking at rows.
    stop_validation_on_column_error: bool = True
    #: A function to materialize a pdt.Table expression returning reference to result.
    # It can also be given a second argument with a table name prefix.
    materialize_hook: Callable[[pdt.Table, str | None], pdt.Table] | None = None

    @classproperty
    def default(cls) -> "Config":
        return Config()


@verb
def alias_subquery(tbl: pdt.Table, cfg: Config, table_prefix: str | None = None):
    if cfg.materialize_hook is not None:
        return cfg.materialize_hook(tbl, table_prefix)
    return pdt.alias(table_prefix)


@verb
def alias_collection_fail(tbl: pdt.Table, cfg: Config, table_prefix: str | None = None):
    if cfg.materialize_hook is not None:
        return cfg.materialize_hook(tbl, table_prefix)
    return pdt.alias(table_prefix)



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import datetime as dt
from collections.abc import Callable

import pydiverse.common as pdc

from ..optional_dependency import ColExpr
from ._base import (
    EPOCH_DATETIME,
    Column,
)
from ._mixins import OrdinalMixin

try:
    import pydiverse.transform as pdt
except ImportError:
    pdt = None

# ------------------------------------------------------------------------------------ #


class Date(OrdinalMixin[dt.date], Column):
    """A column of dates (without time)."""

    def __init__(
        self,
        *,
        nullable: bool = True,
        primary_key: bool = False,
        min: dt.date | None = None,  # noqa: A002
        min_exclusive: dt.date | None = None,
        max: dt.date | None = None,  # noqa: A002
        max_exclusive: dt.date | None = None,
        resolution: str | None = None,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
    ):
        """
        Args:
            nullable: Whether this column may contain null values.
            primary_key: Whether this column is part of the primary key of the schema.
                If ``True``, ``nullable`` is automatically set to ``False``.
            min: The minimum date for dates in this column (inclusive).
            min_exclusive: Like ``min`` but exclusive. May not be specified if ``min``
                is specified and vice versa.
            max: The maximum date for dates in this column (inclusive).
            max_exclusive: Like ``max`` but exclusive. May not be specified if ``max``
                is specified and vice versa.
            resolution: The resolution that dates in the column must have. This uses the
                formatting language used by :mod:`polars` datetime ``round`` method.
                For example, a value ``1mo`` expects all dates to be on the first of the
                month. Note that this setting does *not* affect the storage resolution.
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name.
        """
        # TODO: implement date_matches_resolution
        # if resolution is not None:
        #     offset_time = pl.Series([EPOCH_DATETIME]).dt.offset_by(resolution) \
        #       .dt.time()
        #     if offset_time.item() != dt.time():
        #         raise ValueError("`resolution` is too fine for dates.")
        # if resolution is not None and min is not None:
        #     if not date_matches_resolution(min, resolution):
        #         raise ValueError("`min` does not match resolution.")
        # if resolution is not None and min_exclusive is not None:
        #     if not date_matches_resolution(min_exclusive, resolution):
        #         raise ValueError("`min_exclusive` does not match resolution.")
        # if resolution is not None and max is not None:
        #     if not date_matches_resolution(max, resolution):
        #         raise ValueError("`max` does not match resolution.")
        # if resolution is not None and max_exclusive is not None:
        #     if not date_matches_resolution(max_exclusive, resolution):
        #         raise ValueError("`max_exclusive` does not match resolution.")

        super().__init__(
            nullable=nullable,
            primary_key=primary_key,
            min=min,
            min_exclusive=min_exclusive,
            max=max,
            max_exclusive=max_exclusive,
            check=check,
            alias=alias,
        )
        self.resolution = resolution

    def dtype(self) -> pdc.Date:
        return pdc.Date()

    def validation_rules(self, expr: ColExpr) -> dict[str, ColExpr]:
        result = super().validation_rules(expr)
        # # pydiverse.transform is currently not able to check resolution problems.
        # # However, it is a problem that can be solved by convention.
        # # pydiverse.pipedag will enforce those.
        # if self.resolution is not None:
        #     result["resolution"] = expr.dt.truncate(self.resolution) == expr
        return result


class Time(OrdinalMixin[dt.time], Column):
    """A column of times (without date)."""

    def __init__(
        self,
        *,
        nullable: bool = True,
        primary_key: bool = False,
        min: dt.time | None = None,  # noqa: A002
        min_exclusive: dt.time | None = None,
        max: dt.time | None = None,  # noqa: A002
        max_exclusive: dt.time | None = None,
        resolution: str | None = None,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
    ):
        """
        Args:
            nullable: Whether this column may contain null values.
            primary_key: Whether this column is part of the primary key of the schema.
                If ``True``, ``nullable`` is automatically set to ``False``.
            min: The minimum time for times in this column (inclusive).
            min_exclusive: Like ``min`` but exclusive. May not be specified if ``min``
                is specified and vice versa.
            max: The maximum time for times in this column (inclusive).
            max_exclusive: Like ``max`` but exclusive. May not be specified if ``max``
                is specified and vice versa.
            resolution: The resolution that times in the column must have. This uses the
                formatting language used by :mod:`polars` datetime ``round`` method.
                For example, a value ``1h`` expects all times to be full hours. Note
                that this setting does *not* affect the storage resolution.
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name.
        """
        # # pydiverse.transform is currently not able to check resolution problems.
        # # However, it is a problem that can be solved by convention.
        # # pydiverse.pipedag will enforce those.
        # if resolution is not None:
        #     offset_date = pl.Series([EPOCH_DATETIME]).dt.offset_by(resolution) \
        #       .dt.date()
        #     if offset_date.item() != EPOCH_DATETIME.date():
        #         raise ValueError("`resolution` is too coarse for times.")
        # if resolution is not None and min is not None:
        #     if not time_matches_resolution(min, resolution):
        #         raise ValueError("`min` does not match resolution.")
        # if resolution is not None and min_exclusive is not None:
        #     if not time_matches_resolution(min_exclusive, resolution):
        #         raise ValueError("`min_exclusive` does not match resolution.")
        # if resolution is not None and max is not None:
        #     if not time_matches_resolution(max, resolution):
        #         raise ValueError("`max` does not match resolution.")
        # if resolution is not None and max_exclusive is not None:
        #     if not time_matches_resolution(max_exclusive, resolution):
        #         raise ValueError("`max_exclusive` does not match resolution.")

        super().__init__(
            nullable=nullable,
            primary_key=primary_key,
            min=min,
            min_exclusive=min_exclusive,
            max=max,
            max_exclusive=max_exclusive,
            check=check,
            alias=alias,
        )
        self.resolution = resolution

    def dtype(self) -> pdc.Time:
        return pdc.Time()

    def validation_rules(self, expr: ColExpr) -> dict[str, ColExpr]:
        result = super().validation_rules(expr)
        # # pydiverse.transform is currently not able to check resolution problems.
        # # However, it is a problem that can be solved by convention.
        # # pydiverse.pipedag will enforce those.
        # if self.resolution is not None:
        #     rounded_expr = (
        #         pdt.lit(EPOCH_DATETIME.date())
        #         .dt.combine(expr)
        #         .dt.truncate(self.resolution)
        #         .dt.time()
        #     )
        #     result["resolution"] = rounded_expr == expr
        return result


class Datetime(OrdinalMixin[dt.datetime], Column):
    """A column of datetimes."""

    def __init__(
        self,
        *,
        nullable: bool = True,
        primary_key: bool = False,
        min: dt.datetime | None = None,  # noqa: A002
        min_exclusive: dt.datetime | None = None,
        max: dt.datetime | None = None,  # noqa: A002
        max_exclusive: dt.datetime | None = None,
        resolution: str | None = None,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
    ):
        """
        Args:
            nullable: Whether this column may contain null values.
            primary_key: Whether this column is part of the primary key of the schema.
                If ``True``, ``nullable`` is automatically set to ``False``.
            min: The minimum datetime for datetimes in this column (inclusive).
            min_exclusive: Like ``min`` but exclusive. May not be specified if ``min``
                is specified and vice versa.
            max: The maximum datetime for datetimes in this column (inclusive).
            max_exclusive: Like ``max`` but exclusive. May not be specified if ``max``
                is specified and vice versa.
            resolution: The resolution that datetimes in the column must have. This uses
                the formatting language used by :mod:`polars` datetime ``round`` method.
                For example, a value ``1h`` expects all datetimes to be full hours. Note
                that this setting does *not* affect the storage resolution.
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name.
        """
        # if resolution is not None and min is not None:
        #     if not datetime_matches_resolution(min, resolution):
        #         raise ValueError("`min` does not match resolution.")
        # if resolution is not None and min_exclusive is not None:
        #     if not datetime_matches_resolution(min_exclusive, resolution):
        #         raise ValueError("`min_exclusive` does not match resolution.")
        # if resolution is not None and max is not None:
        #     if not datetime_matches_resolution(max, resolution):
        #         raise ValueError("`max` does not match resolution.")
        # if resolution is not None and max_exclusive is not None:
        #     if not datetime_matches_resolution(max_exclusive, resolution):
        #         raise ValueError("`max_exclusive` does not match resolution.")

        super().__init__(
            nullable=nullable,
            primary_key=primary_key,
            min=min,
            min_exclusive=min_exclusive,
            max=max,
            max_exclusive=max_exclusive,
            check=check,
            alias=alias,
        )
        self.resolution = resolution

    def dtype(self) -> pdc.Datetime:
        return pdc.Datetime()

    def validation_rules(self, expr: ColExpr) -> dict[str, ColExpr]:
        result = super().validation_rules(expr)
        # # pydiverse.transform is currently not able to check resolution problems.
        # # However, it is a problem that can be solved by convention.
        # # pydiverse.pipedag will enforce those.
        # if self.resolution is not None:
        #     result["resolution"] = expr.dt.truncate(self.resolution) == expr
        return result


class Duration(OrdinalMixin[dt.timedelta], Column):
    """A column of durations."""

    def __init__(
        self,
        *,
        nullable: bool = True,
        primary_key: bool = False,
        min: dt.timedelta | None = None,  # noqa: A002
        min_exclusive: dt.timedelta | None = None,
        max: dt.timedelta | None = None,  # noqa: A002
        max_exclusive: dt.timedelta | None = None,
        resolution: str | None = None,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
    ):
        """
        Args:
            nullable: Whether this column may contain null values.
            primary_key: Whether this column is part of the primary key of the schema.
                If ``True``, ``nullable`` is automatically set to ``False``.
            min: The minimum duration for durations in this column (inclusive).
            min_exclusive: Like ``min`` but exclusive. May not be specified if ``min``
                is specified and vice versa.
            max: The maximum duration for durations in this column (inclusive).
            max_exclusive: Like ``max`` but exclusive. May not be specified if ``max``
                is specified and vice versa.
            resolution: The resolution that durations in the column must have. This uses
                the formatting language used by :mod:`polars` datetime ``round`` method.
                For example, a value ``1h`` expects all durations to be full hours. Note
                that this setting does *not* affect the storage resolution.
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name.
        """
        # if resolution is not None and min is not None:
        #     if not timedelta_matches_resolution(min, resolution):
        #         raise ValueError("`min` does not match resolution.")
        # if resolution is not None and min_exclusive is not None:
        #     if not timedelta_matches_resolution(min_exclusive, resolution):
        #         raise ValueError("`min_exclusive` does not match resolution.")
        # if resolution is not None and max is not None:
        #     if not timedelta_matches_resolution(max, resolution):
        #         raise ValueError("`max` does not match resolution.")
        # if resolution is not None and max_exclusive is not None:
        #     if not timedelta_matches_resolution(max_exclusive, resolution):
        #         raise ValueError("`max_exclusive` does not match resolution.")

        super().__init__(
            nullable=nullable,
            primary_key=primary_key,
            min=min,
            min_exclusive=min_exclusive,
            max=max,
            max_exclusive=max_exclusive,
            check=check,
            alias=alias,
        )
        self.resolution = resolution

    def dtype(self) -> pdc.Duration:
        return pdc.Duration()

    def validation_rules(self, expr: ColExpr) -> dict[str, ColExpr]:
        result = super().validation_rules(expr)
        # # pydiverse.transform is currently not able to check resolution problems.
        # # However, it is a problem that can be solved by convention.
        # # pydiverse.pipedag will enforce those.
        # if self.resolution is not None:
        #     datetime = pdt.lit(EPOCH_DATETIME) + expr
        #     result["resolution"] = datetime.dt.truncate(self.resolution) == datetime

        return result


# --------------------------------------- UTILS -------------------------------------- #


def _next_date(t: dt.date, resolution: str | None) -> dt.date | None:
    result = _next_datetime(dt.datetime.combine(t, dt.time()), resolution)
    if result is None:
        return None
    return result.date()


def _next_datetime(t: dt.datetime, resolution: str | None) -> dt.datetime | None:
    from pydiverse.colspec.optional_dependency import pl

    result = pl.Series([t]).dt.offset_by(resolution or "1us")
    if result.dt.year().item() >= 10000:
        # The datetime is out-of-range for a Python datetime object
        return None
    return result.item()


def _next_time(t: dt.time, resolution: str | None) -> dt.time | None:
    from pydiverse.colspec.optional_dependency import pl

    result = pl.cast(
        # `None` can never happen as we can never reach another day by adding time
        dt.datetime,
        _next_datetime(dt.datetime.combine(EPOCH_DATETIME.date(), t), resolution),
    )
    result_time = result.time()
    return None if result_time == dt.time() else result_time


def _next_timedelta(t: dt.timedelta, resolution: str | None) -> dt.timedelta | None:
    from pydiverse.colspec.optional_dependency import pl

    result = pl.cast(
        dt.datetime,  # We run into out-of-date issues before reaching `None`
        _next_datetime(EPOCH_DATETIME + t, resolution),
    )
    return result - EPOCH_DATETIME



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause
import pytest

from pydiverse.colspec import GroupRulePolars, RulePolars
from pydiverse.colspec.optional_dependency import assert_frame_equal, dy, pl
from pydiverse.colspec.testing import evaluate_rules_polars


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_single_column_single_rule():
    lf = pl.LazyFrame({"a": [1, 2]})
    rules = {
        "a|min": RulePolars(pl.col("a") >= 2),
    }
    actual = evaluate_rules_polars(lf, rules)

    expected = pl.LazyFrame({"a|min": [False, True]})
    assert_frame_equal(actual, expected)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_single_column_multi_rule():
    lf = pl.LazyFrame({"a": [1, 2, 3]})
    rules = {
        "a|min": RulePolars(pl.col("a") >= 2),
        "a|max": RulePolars(pl.col("a") <= 2),
    }
    actual = evaluate_rules_polars(lf, rules)

    expected = pl.LazyFrame(
        {"a|min": [False, True, True], "a|max": [True, True, False]}
    )
    assert_frame_equal(actual, expected)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_multi_column_multi_rule():
    lf = pl.LazyFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    rules = {
        "a|min": RulePolars(pl.col("a") >= 2),
        "a|max": RulePolars(pl.col("a") <= 2),
        "b|even": RulePolars(pl.col("b") % 2 == 0),
    }
    actual = evaluate_rules_polars(lf, rules)

    expected = pl.LazyFrame(
        {
            "a|min": [False, True, True],
            "a|max": [True, True, False],
            "b|even": [True, False, True],
        }
    )
    assert_frame_equal(actual, expected)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_cross_column_rule():
    lf = pl.LazyFrame({"a": [1, 1, 2, 2], "b": [1, 1, 1, 2]})
    rules = {"primary_key": RulePolars(~pl.struct("a", "b").is_duplicated())}
    actual = evaluate_rules_polars(lf, rules)

    expected = pl.LazyFrame({"primary_key": [False, False, True, True]})
    assert_frame_equal(actual, expected)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_group_rule():
    lf = pl.LazyFrame({"a": [1, 1, 2, 2, 3], "b": [1, 1, 1, 2, 1]})
    rules: dict[str, RulePolars] = {
        "unique_b": GroupRulePolars(pl.col("b").n_unique() == 1, group_columns=["a"])
    }
    actual = evaluate_rules_polars(lf, rules)

    expected = pl.LazyFrame({"unique_b": [True, True, False, False, True]})
    assert_frame_equal(actual, expected)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_simple_rule_and_group_rule():
    lf = pl.LazyFrame({"a": [1, 1, 2, 2, 3], "b": [1, 1, 1, 2, 1]})
    rules: dict[str, RulePolars] = {
        "b|max": RulePolars(pl.col("b") <= 1),
        "unique_b": GroupRulePolars(pl.col("b").n_unique() == 1, group_columns=["a"]),
    }
    actual = evaluate_rules_polars(lf, rules)

    expected = pl.LazyFrame(
        {
            "b|max": [True, True, True, False, True],
            "unique_b": [True, True, False, False, True],
        }
    )
    assert_frame_equal(actual, expected, check_column_order=False)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_multiple_group_rules():
    lf = pl.LazyFrame({"a": [1, 1, 2, 2, 3], "b": [1, 1, 1, 2, 1]})
    rules: dict[str, RulePolars] = {
        "unique_b": GroupRulePolars(pl.col("b").n_unique() == 1, group_columns=["a"]),
        "sum_b": GroupRulePolars(pl.col("b").sum() >= 2, group_columns=["a"]),
        "group_count": GroupRulePolars(pl.len() >= 2, group_columns=["a", "b"]),
    }
    actual = evaluate_rules_polars(lf, rules)

    expected = pl.LazyFrame(
        {
            "unique_b": [True, True, False, False, True],
            "sum_b": [True, True, True, True, False],
            "group_count": [True, True, False, False, False],
        }
    )
    assert_frame_equal(actual, expected)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.colspec import RulePolars
from pydiverse.colspec.optional_dependency import ColExpr, pdt, pl


def rules_from_exprs_polars(exprs: dict[str, pl.Expr]) -> dict[str, RulePolars]:
    """Turn a set of expressions into simple rules.

    Args:
        exprs: The expressions, mapping from names to :class:`polars.Expr`.

    Returns:
        The rules corresponding to the expressions.
    """
    return {name: RulePolars(expr) for name, expr in exprs.items()}


def evaluate_rules_polars(
    lf: pl.LazyFrame, rules: dict[str, RulePolars]
) -> pl.LazyFrame:
    """Evaluate the provided rules and return the rules' evaluation.

    Args:
        lf: The data frame on which to evaluate the rules.
        rules: The rules to evaluate where the key of the dictionary provides the name
            of the rule.

    Returns:
        The same return value as :meth:`with_evaluation_rules` only that the columns
        of the input data frame are dropped.
    """
    return RulePolars.append_rules_polars(lf, rules).drop(lf.collect_schema())


def evaluate_rules(tbl: pdt.Table, rules: dict[str, ColExpr]):
    return {
        k: (tbl >> pdt.select() >> pdt.mutate(out=v) >> pdt.export(pdt.DictOfLists()))[
            "out"
        ]
        for k, v in rules.items()
    }



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.exc import (
    DtypeValidationError,
    RuleValidationError,
    ValidationError,
)
from pydiverse.colspec.optional_dependency import (
    C,
    ColExpr,
    assert_frame_equal,
    dy,
    pdt,
    pl,
)


class MyColSpec(cs.ColSpec):
    a = cs.Int64(primary_key=True)
    b = cs.String(nullable=False, max_length=5)
    c = cs.String()


class MyComplexColSpec(cs.ColSpec):
    a = cs.Int64()
    b = cs.Int64()

    @cs.rule_polars()
    @staticmethod
    def b_greater_a() -> pl.Expr:
        return pl.col("b") > pl.col("a")

    @cs.rule_polars(group_by=["a"])
    @staticmethod
    def b_unique_within_a() -> pl.Expr:
        return pl.col("b").n_unique() == 1

    @cs.rule()
    @staticmethod
    def b_greater_a2() -> ColExpr:
        return C.b > C.a

    @staticmethod
    @cs.rule(group_by=["a"])
    def b_unique_within_a2() -> ColExpr:
        return C.b.count() == 1  # TODO: n_unique() is not available in pdt


# -------------------------------------- COLUMNS ------------------------------------- #


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_missing_columns(df_type: type[pl.DataFrame] | type[pl.LazyFrame]):
    df = df_type({"a": [1], "b": [""]})
    with pytest.raises(ValidationError):
        MyColSpec.validate_polars(df)
    assert not MyColSpec.is_valid_polars(df)


# -------------------------------------- DTYPES -------------------------------------- #


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_invalid_dtype(df_type: type[pl.DataFrame] | type[pl.LazyFrame]):
    df = df_type({"a": [1], "b": [1], "c": [1]})
    try:
        MyColSpec.validate_polars(df)
        raise AssertionError()  # above should raise
    except DtypeValidationError as exc:
        assert len(exc.errors) == 2
    assert not MyColSpec.is_valid_polars(df)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_invalid_dtype_cast(df_type: type[pl.DataFrame] | type[pl.LazyFrame]):
    df = df_type({"a": [1], "b": [1], "c": [1]})
    actual = MyColSpec.validate_polars(df, cast=True)
    expected = pl.DataFrame({"a": [1], "b": ["1"], "c": ["1"]})
    assert_frame_equal(actual, expected)
    assert MyColSpec.is_valid_polars(df, cast=True)


# --------------------------------------- RULES -------------------------------------- #


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_invalid_column_contents(df_type: type[pl.DataFrame] | type[pl.LazyFrame]):
    df = df_type({"a": [1, 2, 3], "b": ["x", "longtext", None], "c": ["1", None, "3"]})
    try:
        MyColSpec.validate_polars(df)
        raise AssertionError()  # above should raise
    except RuleValidationError as exc:
        assert len(exc.schema_errors) == 0
        assert exc.column_errors == {"b": {"nullability": 1, "max_length": 1}}
    assert not MyColSpec.is_valid_polars(df)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_invalid_primary_key(df_type: type[pl.DataFrame] | type[pl.LazyFrame]):
    df = df_type({"a": [1, 1], "b": ["x", "y"], "c": ["1", "2"]})
    try:
        MyColSpec.validate_polars(df)
        raise AssertionError()  # above should raise
    except RuleValidationError as exc:
        assert exc.schema_errors == {"primary_key": 2}
        assert len(exc.column_errors) == 0
    assert not MyColSpec.is_valid_polars(df)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_violated_custom_rule_polars(df_type: type[pl.DataFrame] | type[pl.LazyFrame]):
    df = df_type({"a": [1, 1, 2, 3, 3], "b": [2, 2, 2, 4, 5]})
    try:
        MyComplexColSpec.validate_polars(df)
        raise AssertionError()  # above should raise
    except RuleValidationError as exc:
        assert exc.schema_errors == {"b_greater_a": 1, "b_unique_within_a": 2}
        assert len(exc.column_errors) == 0
    assert not MyComplexColSpec.is_valid_polars(df)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_violated_custom_rule():
    tbl = pdt.Table({"a": [1, 1, 2, 3, 3], "b": [2, 2, 2, 4, 5]})
    try:
        MyComplexColSpec.validate(tbl)
        raise AssertionError()  # above should raise
    except RuleValidationError as exc:
        assert exc.schema_errors == {"b_greater_a2": 1, "b_unique_within_a2": 4}
        assert len(exc.column_errors) == 0
    assert not MyComplexColSpec.is_valid(tbl)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("df_type", [pl.DataFrame, pl.LazyFrame])
def test_success_multi_row_strip_cast(
    df_type: type[pl.DataFrame] | type[pl.LazyFrame],
):
    df = df_type(
        {"a": [1, 2, 3], "b": ["x", "y", "z"], "c": [1, None, None], "d": [1, 2, 3]}
    )
    actual = MyColSpec.validate_polars(df, cast=True)
    expected = pl.DataFrame(
        {"a": [1, 2, 3], "b": ["x", "y", "z"], "c": ["1", None, None]}
    )
    assert_frame_equal(actual, expected)
    assert MyColSpec.is_valid_polars(df, cast=True)



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Callable
from pathlib import Path

import pytest

import pydiverse.colspec as cs
import pydiverse.colspec.collection
from pydiverse.colspec.colspec import ColSpecMeta, convert_to_dy_col_spec, dy
from pydiverse.colspec.exc import ImplementationError
from pydiverse.colspec.optional_dependency import SchemaMeta, assert_frame_equal, pl


class MyFirstColSpec(cs.ColSpec):
    a = cs.UInt8(primary_key=True)


class MySecondColSpec(cs.ColSpec):
    a = cs.UInt16(primary_key=True)
    b = cs.Integer


class MyCollection(pydiverse.colspec.collection.Collection):
    first: MyFirstColSpec
    second: MySecondColSpec | None


def test_common_primary_keys():
    assert MyCollection.common_primary_keys() == ["a"]


def test_members():
    members = MyCollection.members()
    assert not members["first"].is_optional
    assert members["second"].is_optional


def test_member_col_specs():
    schemas = MyCollection.member_col_specs()
    assert schemas == {"first": MyFirstColSpec, "second": MySecondColSpec}


def test_required_members():
    required_members = MyCollection.required_members()
    assert required_members == {"first"}


def test_optional_members():
    optional_members = MyCollection.optional_members()
    assert optional_members == {"second"}


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_cast():
    collection = MyCollection.cast_polars_data(
        {
            "first": pl.LazyFrame({"a": [1, 2, 3]}),
            "second": pl.LazyFrame({"a": [1, 2, 3], "b": [4, 5, 6]}),
        },
    )
    assert (
        collection.first.collect_schema()
        == MyFirstColSpec.create_empty_polars().collect_schema()
    )
    assert collection.second is not None
    assert (
        collection.second.collect_schema()
        == MySecondColSpec.create_empty_polars().collect_schema()
    )


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_cast2():
    collection = MyCollection._init_polars_data(
        {
            "first": pl.LazyFrame({"a": [1, 2, 3]}),
            "second": pl.LazyFrame({"a": [1, 2, 3], "b": [4, 5, 6]}),
        },
    )
    collection = collection.cast_polars()
    assert (
        collection.first.collect_schema()
        == MyFirstColSpec.create_empty_polars().collect_schema()
    )
    assert collection.second is not None
    assert (
        collection.second.collect_schema()
        == MySecondColSpec.create_empty_polars().collect_schema()
    )


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    "expected",
    [
        {
            "first": pl.LazyFrame({"a": [1, 2, 3]}, schema={"a": pl.UInt8}),
            "second": pl.LazyFrame(
                {"a": [1, 2, 3], "b": [4, 5, 6]}, schema={"a": pl.UInt16, "b": pl.Int64}
            ),
        },
        {"first": pl.LazyFrame({"a": [1, 2, 3]}, schema={"a": pl.UInt8})},
    ],
)
def test_to_dict(expected: dict[str, pl.LazyFrame]):
    collection = MyCollection.validate_polars_data(expected)

    # Check that export looks as expected
    observed = collection.to_dict()
    assert set(expected.keys()) == set(observed.keys())
    for key in expected.keys():
        pl.testing.assert_frame_equal(expected[key], observed[key])

    # Make sure that "roundtrip" validation works
    assert MyCollection.is_valid_polars_data(observed)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_collect_all():
    collection = MyCollection.cast_polars_data(
        {
            "first": pl.LazyFrame({"a": [1, 2, 3]}).filter(pl.col("a") < 3),
            "second": pl.LazyFrame({"a": [1, 2, 3], "b": [4, 5, 6]}).filter(
                pl.col("b") <= 5
            ),
        }
    )
    out = collection.collect_all_polars()

    assert isinstance(out, MyCollection)
    assert out.first.explain() == 'DF ["a"]; PROJECT */1 COLUMNS'
    assert len(out.first.collect()) == 2
    assert out.second is not None
    assert out.second.explain() == 'DF ["a", "b"]; PROJECT */2 COLUMNS'
    assert len(out.second.collect()) == 2


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_collect_all_optional():
    collection = MyCollection.cast_polars_data(
        {"first": pl.LazyFrame({"a": [1, 2, 3]})}
    )
    out = collection.collect_all_polars()

    assert isinstance(out, MyCollection)
    assert len(out.first.collect()) == 3
    assert out.second is None


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    "read_fn", [MyCollection.scan_parquet, MyCollection.read_parquet]
)
def test_read_write_parquet(tmp_path: Path, read_fn: Callable[[Path], MyCollection]):
    collection = MyCollection.cast_polars_data(
        {
            "first": pl.LazyFrame({"a": [1, 2, 3]}),
            "second": pl.LazyFrame({"a": [1, 2], "b": [10, 15]}),
        }
    )
    collection.write_parquet(tmp_path)

    read = read_fn(tmp_path)
    assert_frame_equal(collection.first, read.first)
    assert collection.second is not None
    assert read.second is not None
    assert_frame_equal(collection.second, read.second)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    "read_fn", [MyCollection.scan_parquet, MyCollection.read_parquet]
)
def test_read_write_parquet_optional(
    tmp_path: Path, read_fn: Callable[[Path], MyCollection]
):
    collection = MyCollection.cast_polars_data(
        {"first": pl.LazyFrame({"a": [1, 2, 3]})}
    )
    collection.write_parquet(tmp_path)

    read = read_fn(tmp_path)
    assert_frame_equal(collection.first, read.first)
    assert collection.second is None
    assert read.second is None


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_dataframely_columns_fail():
    class FailColSpec(cs.ColSpec):
        a = cs.Float64()
        b = dy.String()

    class FailColSpec2(cs.ColSpec):
        a = cs.Float64()
        b = dy.String

    with pytest.raises(
        ImplementationError, match="Dataframely Columns won't work in ColSpec classes."
    ):
        FailColSpec.column_names()

    with pytest.raises(
        ImplementationError, match="Dataframely Columns won't work in ColSpec classes."
    ):
        FailColSpec2.column_names()

    class FailCollection(cs.Collection):
        first: FailColSpec

    class FailCollection2(cs.Collection):
        first: FailColSpec2

    with pytest.raises(
        ImplementationError, match="Dataframely Columns won't work in ColSpec classes."
    ):
        cs.collection.convert_collection_to_dy(FailCollection)
    with pytest.raises(
        ImplementationError, match="Dataframely Columns won't work in ColSpec classes."
    ):
        cs.collection.convert_collection_to_dy(FailCollection2)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_dataframely_columns_fail_inheritance():
    class GoodColSpec(cs.ColSpec):
        a = cs.Float64()

    class SpecialMeta(ColSpecMeta, SchemaMeta):
        pass

    class GoodSchema(dy.Schema, metaclass=SpecialMeta):
        b = dy.Float64()

    assert GoodColSpec.column_names() == ["a"]
    assert GoodSchema.column_names() == ["b"]

    class FailColSpec(GoodColSpec, GoodSchema):
        pass

    with pytest.raises(
        ImplementationError,
        match="Don't mix Dataframely Schema with ColSpec classes in inheritance",
    ):
        convert_to_dy_col_spec(FailColSpec)

    with pytest.raises(
        ImplementationError, match="Dataframely Columns won't work in ColSpec classes."
    ):
        FailColSpec.column_names()


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_dataframely_conversion_success():
    class GoodColSpec(cs.ColSpec):
        a = cs.Float64()

    assert GoodColSpec.column_names() == ["a"]

    class GoodCollection(cs.Collection):
        first: GoodColSpec

    coll = cs.collection.convert_collection_to_dy(GoodCollection)
    assert coll.members()["first"].schema.column_names() == ["a"]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any

import pytest

import pydiverse.colspec as cs
import pydiverse.colspec.collection
from pydiverse.colspec.columns import ColExpr
from pydiverse.colspec.optional_dependency import C, dy, pdt, pl
from pydiverse.colspec.pdt_util import num_rows


class CarColSpec(cs.ColSpec):
    vin = cs.String(primary_key=True)
    manufacturer = cs.String(nullable=False)

    @cs.filter()
    def not_empty(self) -> ColExpr:
        return self.manufacturer != ""


class CarPartColSpec(cs.ColSpec):
    vin = cs.String(primary_key=True)
    part = cs.String(primary_key=True)
    price = cs.Float64(primary_key=True)


class CarFleetPolars(pydiverse.colspec.collection.Collection):
    cars: CarColSpec
    car_parts: CarPartColSpec

    @cs.filter_polars()
    def not_car_with_vin_123(self) -> pl.LazyFrame:
        return self.cars.filter(pl.col("vin") != pl.lit("123"))


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_valid_failure_infos_polars():
    cars = {"vin": ["123", "456"], "manufacturer": ["BMW", "Mercedes"]}
    car_parts: dict[str, list[Any]] = {
        "vin": ["123", "123", "456"],
        "part": ["Motor", "Wheel", "Motor"],
        "price": [1000, 100, 1000],
    }
    car_fleet, failures = CarFleetPolars.filter_polars_data(
        {"cars": pl.DataFrame(cars), "car_parts": pl.DataFrame(car_parts)},
        cast=True,
    )

    assert len(car_fleet.cars.collect()) + len(failures["cars"].invalid()) == len(
        cars["vin"]
    )
    assert len(car_fleet.car_parts.collect()) + len(
        failures["car_parts"].invalid()
    ) == len(car_parts["vin"])
    assert len(failures["cars"].invalid()) == 1
    assert len(failures["car_parts"].invalid()) == 2


@dataclass
class CarFleet(pydiverse.colspec.collection.Collection):
    cars: CarColSpec
    car_parts: CarPartColSpec

    @cs.filter()
    def not_car_with_vin_123(self) -> ColExpr:
        return self.cars.vin != "123"


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_valid_failure_infos():
    cars = {"vin": ["123", "456"], "manufacturer": ["BMW", "Mercedes"]}
    car_parts: dict[str, list[Any]] = {
        "vin": ["123", "123", "456"],
        "part": ["Motor", "Wheel", "Motor"],
        "price": [1000, 100, 1000],
    }
    raw_fleet = CarFleet.build()  # type: CarFleet[pdt.Table]
    raw_fleet.cars = pdt.Table(cars)
    raw_fleet.car_parts = pdt.Table(car_parts)

    car_fleet, failures = raw_fleet.filter(cast=True)  # type: CarFleet[pdt.Table], CarFleet[FailureInfo]

    assert num_rows(car_fleet.cars) + num_rows(failures.cars.invalid_rows) == len(cars)
    assert num_rows(car_fleet.car_parts) + num_rows(
        failures.car_parts.invalid_rows
    ) == len(car_parts)
    assert num_rows(failures.cars.invalid_rows) == 1
    assert num_rows(failures.car_parts.invalid_rows) == 2



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import pl
from pydiverse.colspec.testing.factory import create_colspec


@pytest.mark.skipif(pl.Expr is None, reason="polars is required for this test")
def test_polars_schema() -> None:
    schema = create_colspec("test", {"a": cs.Int32(nullable=False), "b": cs.Float32()})
    pl_schema = schema.polars_schema()
    assert pl_schema == {"a": pl.Int32, "b": pl.Float32}



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

from pydiverse.colspec import Column
from pydiverse.colspec.optional_dependency import C, pdt, pl
from pydiverse.colspec.testing import COLUMN_TYPES, SUPERTYPE_COLUMN_TYPES
from pydiverse.colspec.testing.rules import evaluate_rules


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("column_type", COLUMN_TYPES + SUPERTYPE_COLUMN_TYPES)
@pytest.mark.parametrize("nullable", [True, False])
def test_rule_count_nullability(column_type: type[Column], nullable: bool):
    column = column_type(nullable=nullable)
    assert len(column.validation_rules(C.a)) == int(not nullable)


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("column_type", COLUMN_TYPES + SUPERTYPE_COLUMN_TYPES)
def test_nullability_rule_for_primary_key(column_type: type[Column]):
    column = column_type(primary_key=True)
    assert len(column.validation_rules(C.a)) == 1


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize("column_type", COLUMN_TYPES + SUPERTYPE_COLUMN_TYPES)
def test_nullability_rule(column_type: type[Column]):
    column = column_type(nullable=False)
    dtype = column.dtype()
    polars_type = dtype.to_polars()
    lf = pl.LazyFrame({"a": [None]}, schema={"a": polars_type})
    tbl = pdt.Table(lf)
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    expected = {"nullability": [False]}
    assert actual == expected



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.columns._base import Column
from pydiverse.colspec.optional_dependency import dy, pl
from pydiverse.colspec.testing.factory import create_colspec


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    ("column", "dtype"),
    [
        (cs.Any(), pl.Null()),
        (cs.Bool(), pl.Boolean()),
        (cs.Date(), pl.Date()),
        (cs.Datetime(), pl.Datetime()),
        (cs.Time(), pl.Time()),
        (cs.Duration(), pl.Duration()),
        (cs.Decimal(), pl.Decimal()),
        (cs.Decimal(12), pl.Decimal(12)),
        (cs.Decimal(None, 8), pl.Decimal(None, 8)),
        (cs.Decimal(6, 2), pl.Decimal(6, 2)),
        (cs.Float(), pl.Float64()),
        (cs.Float32(), pl.Float32()),
        (cs.Float64(), pl.Float64()),
        (cs.Integer(), pl.Int64()),
        (cs.Int8(), pl.Int8()),
        (cs.Int16(), pl.Int16()),
        (cs.Int32(), pl.Int32()),
        (cs.Int64(), pl.Int64()),
        (cs.UInt8(), pl.UInt8()),
        (cs.UInt16(), pl.UInt16()),
        (cs.UInt32(), pl.UInt32()),
        (cs.UInt64(), pl.UInt64()),
        (cs.String(), pl.String()),
        (cs.List(cs.String()), pl.List(pl.String())),
        (cs.Struct({"a": cs.String()}), pl.Struct({"a": pl.String()})),
        (cs.Enum(["a", "b"]), pl.Enum(["a", "b"])),
    ],
)
def test_default_dtype(column: Column, dtype: pl.DataType):
    schema = create_colspec("test", {"a": column})
    df = schema.create_empty_polars()
    assert df.schema["a"] == dtype
    schema.validate_polars(df)
    assert schema.is_valid_polars(df)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Callable
from typing import Generic, TypeVar

from .optional_dependency import ColExpr, pl

C = TypeVar("C")


class Filter(Generic[C]):
    """Internal class representing logic for filtering members of a collection."""

    def __init__(self, logic_fn: Callable[[C], ColExpr]):
        self.logic_fn: Callable[[C], ColExpr] = logic_fn


def filter() -> Callable[[Callable[[C], ColExpr]], Filter[C]]:  # noqa: A001
    """Mark a function as filters for rows in the members of a collection.

    The name of the function will be used as the name of the filter. The name must not
    clash with the name of any column in the member schemas or rules defined on the
    member schemas.

    A filter receives a collection as input and must return a data frame like the
    following:

    - The columns must be a superset of the common primary keys across all members.
    - The rows must provide the primary keys which ought to be *kept* across the
      members. The filter results in the removal of rows which are lost as the result
      of inner-joining members onto the return value of this function.

    Attention:
        Make sure to provide unique combinations of the primary keys or the filters
        might introduce duplicate rows.
    """

    def decorator(pred: Callable[[C], ColExpr]) -> Filter[C]:
        return Filter(logic_fn=pred)

    return decorator


class FilterPolars(Generic[C]):
    """Internal class representing logic for filtering members of a collection."""

    def __init__(self, logic: Callable[[C], pl.LazyFrame]):
        self.logic = logic


def filter_polars() -> Callable[[Callable[[C], pl.LazyFrame]], FilterPolars[C]]:
    """Mark a function as filters for rows in the members of a collection.

    The name of the function will be used as the name of the filter. The name must not
    clash with the name of any column in the member schemas or rules defined on the
    member schemas.

    A filter receives a collection as input and must return a data frame like the
    following:

    - The columns must be a superset of the common primary keys across all members.
    - The rows must provide the primary keys which ought to be *kept* across the
      members. The filter results in the removal of rows which are lost as the result
      of inner-joining members onto the return value of this function.

    Attention:
        Make sure to provide unique combinations of the primary keys or the filters
        might introduce duplicate rows.
    """

    def decorator(validation_fn: Callable[[C], pl.LazyFrame]) -> FilterPolars[C]:
        return FilterPolars(logic=validation_fn)

    return decorator



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec import Rule
from pydiverse.colspec.exc import ImplementationError
from pydiverse.colspec.optional_dependency import C, pdt
from pydiverse.colspec.testing.factory import create_colspec


class MyColSpec(cs.ColSpec):
    a = cs.Integer(primary_key=True)
    b = cs.String(primary_key=True)
    c = cs.Float64()
    e = cs.Any()


def test_column_names():
    _ = MyColSpec.e
    with pytest.raises(AttributeError):
        _ = MyColSpec.d
    assert MyColSpec.column_names() == ["a", "b", "c", "e"]


def test_columns():
    columns = MyColSpec.columns()
    assert isinstance(columns["a"], cs.Integer)
    assert isinstance(columns["b"], cs.String)
    assert isinstance(columns["c"], cs.Float64)
    assert isinstance(columns["e"], cs.Any)


def test_nullability():
    columns = MyColSpec.columns()
    assert not columns["a"].nullable
    assert not columns["b"].nullable
    assert columns["c"].nullable
    assert columns["e"].nullable


def test_primary_keys():
    assert MyColSpec.primary_keys() == ["a", "b"]


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_no_rule_named_primary_key():
    tbl = pdt.Table(dict(a=["b", "c"]))
    with pytest.raises(
        ImplementationError,
        match="@cs.rule annotated functions must not be called `_primary_key_`",
    ):
        create_colspec(
            "test",
            {"a": cs.String()},
            {"_primary_key_": Rule(tbl.a.str.len() > 1)},
        ).validate(tbl)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from .optional_dependency import pdt


def num_rows(tbl: pdt.Table) -> int:
    return tbl >> pdt.summarize(num_rows=pdt.count()) >> pdt.export(pdt.Scalar)
